#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
listManipulator.py
===================

Contains a set of list-related functions and the RecursiveListManipulator 
class that is wrapped in nice functions. This class transforms
arrays into multi-dimensional list objects. Wrapper functions are
provided to call these directly as functions.

"""

# Import local modules
import nappy.utils

# Define module variables
default_delimiter = nappy.utils.getDefault("default_delimiter")
default_float_format = nappy.utils.getDefault("default_float_format")


def arrayToList(array, inlist):
    """ 
    Takes an n-dimensional Numeric array and converts it to an 
    n-dimensional list object.
    """
    dimlist = array.shape
    if len(dimlist[1:]) > 0:
        for i in range(dimlist[0]):
            arrayToList(inlist[i], array[i])
    else:
        for i in range(dimlist[0]):
            inlist.append(array[i])
    return inlist


def listOfListsCreator(inlist, dimlist):
    """
    Creates a list of lists with dimensions defined in dimlist (a list of integers).
    """
    if len(dimlist[1:]) > 0:
        for i in range(dimlist[0]):
            inlist.append([])
            listOfListsCreator(inlist[i], array, dimlist[1:])
    return inlist


def recursiveListPopulator(inlist, array, dimlist):
    "Function wrapper around class method RecursiveListManipulator().populator()."
    return RecursiveListManipulator().populator(inlist, array, dimlist)


def recursiveListWriter(inlist, dimlist, delimiter=default_delimiter, float_format=default_float_format):
    "Function wrapper around class method RecursiveListManipulator().writeLines()."
    return RecursiveListManipulator().writeLines(inlist, dimlist, delimiter=delimiter, float_format=float_format)


class RecursiveListManipulator:
    """
    Container class with methods to convert a 1-D array into a multi-dimensional list.
    """

    def populator(self, inlist, array, dimlist):    
        """
        Populates the list object 'inlist' (e.g. []) with sublists of
        dimensionality defined in the 'dimlist' list of dimensions (e.g [181, 360]).
        At the deepest level it then inserts values from the long list
        'array'.
        """

        if not hasattr(self, "_counter"):
            self._counter = 0
        if len(dimlist[1:]) > 0:
            for i in range(dimlist[0]):
                inlist.append([])
                self.populator(inlist[i], array, dimlist[1:])
        else:
            count = self._counter
            self._counter = self._counter + dimlist[0]
            endcount = self._counter
            for i in range(count, endcount):
                inlist.append(array[i])
        return inlist

    def writeLines(self, inlist, dimlist, delimiter=default_delimiter, float_format=default_float_format):
        """
        Method to walk through all the levels of the multi-level list object
        'inlist' and writes out appropriate values to a list of lines called
        'self.rtlines'. 'dimlist' is a list of the dimensions within 'inlist'.
        """
        if not hasattr(self, "rtlines"):
            self.rtlines = []
        if len(dimlist[1:]) > 0:
            for i in range(dimlist[0]):
                self.writeLines(inlist[i], dimlist[1:], delimiter=delimiter, float_format=float_format)
        else:
            var_string = ""
            for i in range(dimlist[0]):
                var_string = var_string + ((float_format+delimiter) % inlist[i])
            self.rtlines.append("%s\n" % var_string.rstrip(" ,"))
        return self.rtlines
            



#!/usr/bin/env python

"""
compare.py
==========

Tool to compare contents of files or directories full of files.

Usage:
======

    compare.py <item1> <item2>

Where:
======

    <item1> and <item2> can either be a text file or directory.
"""

# Import standard library modules
import os
import sys
import re

file_exclusion_patterns = (".*CSV.*", ".*svn.*", "\..*", ".*\.pyc$", ".*~$") 
file_exclusions = [re.compile(pattn) for pattn in file_exclusion_patterns]
dir_exclusion_patterns = (".*CSV.*", ".*svn.*")
dir_exclusions = [re.compile(pattn) for pattn in dir_exclusion_patterns]


def exitNicely(msg):
    "Tidy exit."
    print __doc__
    print msg
    sys.exit()


def compare(i1, i2):
    """
    Compares items whether files or directories.
    Reports any differences at the command line but
    also returns them in a dictionary as:
    ???
    """
    if os.path.isfile(i1):
        compFiles(i1, i2)
    elif os.path.isdir(i1):
        compDirs(i1, i2)
    else:
        exitNicely("Cannot recognise/find item '" + i1 + "'.")


def compDirs(d1, d2):
    """
    Compares directories by looping through and then calling
    compFiles() for each pair of files found.
    """
    dname = os.path.split(d1)[-1]
    # Ignore anything that is in exclusion list
    for excl in dir_exclusions:
        if excl.match(dname):
            print "IGNORING EXCLUDED Directory:", d1
            return

    items = os.listdir(d1)
    
    for item in items:

        d1f = os.path.join(d1, item)
	d2f = os.path.join(d2, item)
	
        if not os.path.exists(d2f):
            print "WARNING: cannot find item:", d2f
            continue

	if os.path.isdir(d1f): 
	    compDirs(d1f, d2f)
	    continue

        compFiles(d1f, d2f)


def compFiles(f1, f2):
    """
    Compares contents of two files.
    """	
    name = os.path.split(f1)[-1]
    # Ignore anything that is in exclusion list
    for excl in file_exclusions:
        if excl.match(name):
            print "IGNORING EXCLUDED file:", f1
            return

    # Check they exist
    for f in (f1, f2):
        if not os.path.isfile(f):
            exitNicely("CANNOT compare files as item does not exist:" + f)
            

    l1 = open(f1).readlines()
    l2 = open(f2).readlines()

    leng = len(l1)
    if len(l2) < leng: 
        leng=len(l2)

    print "\n>>>", f1, "\n<<<", f2

    for i in range(leng):
        if l1[i] != l2[i]:
            print "Line %s:" % (i+1)
            print ">>>", l1[i]
            print "<<<", l2[i]

    
if __name__=="__main__":

    args = sys.argv[1:]
    if len(args) != 2:
       exitNicely("Must provide two items to compare as command-line arguments.")

    compare(args[0], args[1])



#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naFile4010.py
=============

Container module for NAFile4010 class.

"""

# Imports from python standard library

# Imports from local package
import nappy.na_file.na_file_2010

class NAFile4010(nappy.na_file.na_file_2010.NAFile2010):
    """
    Class to read, write and interact with NASA Ames files conforming to the
    File Format Index (FFI) 4010.
    
    Identical class rules to FFI 2010. Sub-classed for neatness.
    """
    pass



#!/usr/bin/env python
#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
aircraft_utils.py
==================

UNSUPPORTED!

"""

# Imports from python standard library

# Import external packages
import cdms
import Numeric
import MV

cdms.setAutoBounds("off") 


def flatten2DTimeData(var, time_var):
    """
    Returns a flattened 2D array variable with a recalculated time axies.
    """
    data = MV.ravel(var)
    missing = var.getMissing()
    time_values = time_var._data
    time_unit = time_var.units
    sampling_rate = var.getAxis(1)
    rate = int(sampling_rate.id[3:])
    newtime_values = []

    for t in time_values:
        for i in range(rate):
            tvalue = t + ((1./rate)*i)
            newtime_values.append(tvalue)

    new_time_ax = cdms.createAxis(newtime_values)
    new_time_ax.units = time_unit
    new_time_ax.designateTime()
    new_time_ax.id = new_time_ax.long_name = new_time_ax.standard_name = "time"

    atts = var.attributes
    newAtts = {}
    for att,value in atts.items():
        if type(value) in (type((1,1)), type([1,2]), type(Numeric.array([1.]))) and len(value) == 1:
            value = value[0]
        if type(value) in (type(1), type(1.), type(long(1))):
            newAtts[att] = value
        elif type(value) == type("string"):
            newAtts[att] = value.strip()

    # Now create the variable
    newvar = cdms.createVariable(data, id=var.id, fill_value=missing, axes=[new_time_ax], attributes=newAtts)
    return newvar


#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naError.py
==========

Container module for nappy error definitions.

"""

# Define global variables
NAPlatformError="Platform error: You cannot use the CDAT-dependent features on a Windows Platform."


#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naFile1001.py
=============

Container module for NAFile1001 class.

"""

# Imports from python standard library

# Imports from local package
import nappy.utils.text_parser
import nappy.na_file.na_file
import nappy.utils.common_utils
getAnnotation = nappy.utils.common_utils.getAnnotation
wrapLine = nappy.utils.common_utils.annotateLine
wrapLines = nappy.utils.common_utils.annotateLines


class NAFile1001(nappy.na_file.na_file.NAFile):
    """
    Class to read, write and interact with NASA Ames files conforming to the
    File Format Index (FFI) 1001.
    """

    def readHeader(self):
        """
        Reads FFI-specific header section.
        """
        self._readCommonHeader()
        self.DX = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), self.NIV, float)
        self.XNAME = nappy.utils.text_parser.readItemsFromLines(self._readLines(self.NIV), self.NIV, str)
        self._readVariablesHeaderSection()
        self._readComments()

    def writeHeader(self):
        """
        Writes FFI-specific header section.
        """        
        self._writeCommonHeader()
        annotation = getAnnotation("DX", self.annotation, delimiter = self.delimiter)
        self.header.write((annotation + "%s" * self.NIV + "\n") % tuple(self.DX))
        annotation = getAnnotation("XNAME", self.annotation, delimiter = self.delimiter)
        self.header.write((annotation + "%s\n") * self.NIV % tuple(self.XNAME))
        self._writeVariablesHeaderSection()
        self._writeComments()
        self._fixHeaderLength()
        self.file.write(self.header.read())

    def _setupArrays(self):
        """
        Sets up FFI-specific arrays to fill with data (lists of lists).
        """
        self.X = []
        self.V = []
        # Set up the variables list
        for n in range(self.NV):
            self.V.append([])

    def _readData1(self, datalines, ivar_count):
        """
        Reads first line/section of current block of data.
        """
        (x_and_v, rtlines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, 1 + self.NV, float)
        (x, v) = (x_and_v[0], x_and_v[1:])
        self.X.append(x)
        count = 0
        # Set up mth list in self.V
        for n in range(self.NV):
            self.V[n].append(v[count])
            count = count + 1
        return rtlines    

    def _readData2(self, datalines, ivar_count):
        """
        Reads second line/section (if used) of current block of data.
        """
        return datalines  

    def writeData(self):
        """
        Writes the data section of the file.
        This method can be called directly by the user.
        """
        for m in range(len(self.X)):

            var_string = self.format % self.X[m]

            for n in range(self.NV):
                var_string = var_string + (self.format % self.V[n][m])

            self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(" ,")))



#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naFile2110.py
=============

Container module for NAFile2110 class.

"""

# Imports from python standard library

# Imports from local package
import nappy.utils.text_parser
import nappy.na_file.na_file_2010
import nappy.utils.common_utils
wrapLine = nappy.utils.common_utils.annotateLine
wrapLines = nappy.utils.common_utils.annotateLines

class NAFile2110(nappy.na_file.na_file_2010.NAFile2010):
    """
    Class to read, write and interact with NASA Ames files conforming to the
    File Format Index (FFI) 2110.
    """

    def readHeader(self):
        """
        Reads FFI-specifc header section.
        """
        self._readCommonHeader()
        self.DX = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), self.NIV, float)
        self.DX.reverse()  # Reverse because C-type array is least-changing first
        self.XNAME = nappy.utils.text_parser.readItemsFromLines(self._readLines(self.NIV), self.NIV, str)
        self.XNAME.reverse()  # Reverse because C-type array is least-changing first
        self._readVariablesHeaderSection()
        self._readAuxVariablesHeaderSection()
        self._readComments()

    def writeHeader(self):
        """
        Writes FFI-specific header section.
        """
        self._writeCommonHeader()
        DX = self.DX
        DX.reverse()
        self.header.write(wrapLine("DX", self.annotation, self.delimiter, (("%s" + self.delimiter) * (self.NIV - 1) + "%s\n") % tuple(DX)))
        XNAME = self.XNAME
        XNAME.reverse()
        self.header.write(wrapLines("XNAME", self.annotation, self.delimiter, "%s\n" * self.NIV % tuple(XNAME)))
        self._writeVariablesHeaderSection()
        self._writeAuxVariablesHeaderSection()
        self._writeComments()
        self._fixHeaderLength()
        self.file.write(self.header.read())


    def _setupArrays(self):
        """
        Sets up FFI-specific arrays to fill with data (lists of lists).
        """
        self.V = []
        self.A = []
        self.X = []  # Needs to be a list of sublists each containing [x0n, [x1n, x1n + 1, x1n + 2....]]
        self.NX = []

        for n in range(self.NV):
            self.V.append([])
        for i in range(self.NAUXV):
           self.A.append([])

    def _readData1(self, datalines, ivar_count): 
        """
        Reads first line/section of current block of data.
        """    
        # Start with independent and Auxilliary vars
        (x_and_a, rtlines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, self.NAUXV + 1, float)
        (x, aux) = (x_and_a[0], x_and_a[1:])
	count = 0
        for a in range(self.NAUXV):
            self.A[a].append(aux[count])
            count = count + 1
        #for a in range(self.NAUXV):
            #self.A.append(aux[a])
        self.X.append([])
        self.X[ivar_count].append(x)
        # Set up list to take second changing independent variable
        self.X[ivar_count].append([])  
        self.NX.append(int(aux[0]))
        return rtlines

    def _readData2(self, datalines, ivar_count):
        """
        Reads second line/section (if used) of current block of data.
        """
        # Now get the dependent variables
        for n in range(self.NV):
            self.V[n].append([])

        for c in range(self.NX[ivar_count]):
            (x_and_v, datalines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, self.NV + 1, float)
            (x, v) = (x_and_v[0], x_and_v[1:])
            self.X[ivar_count][1].append(x)

            count = 0
            for n in range(self.NV):
                self.V[n][ivar_count].append(v[count])
                count = count + 1

        rtlines = datalines
        return rtlines

    def writeData(self):
        """
        Writes the data section of the file.
        This method can be called directly by the user.
        """
        # Set up unbounded IV loop

        for m in range(len(self.X)):

            # Write unbounded independent variable mark and auxiliary variables
            var_string = self.format % self.X[m][0]

            # Loop through aux vars which includes NX as first aux var
            for a in range(self.NAUXV): 
                var_string = var_string + (self.format % self.A[a][m])
            self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(" ,")))

            # Write second independant variable and dependant variables
            for p in range(self.NX[m]):
                var_string = self.format % self.X[m][1][p]

                for n in range(self.NV):
                    var_string = var_string + (self.format %self.V[n][m][p])

                self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" %var_string.rstrip(" ,")))



#!/usr/bin/env python

"""
scanFAAM.py
===========

Holds the scanFAAM function that is used to work out the first and last
index of real data (i.e. non-missing data in a load of FAAM flight data).
It is used in conjunction with nappy to reduce the size of output files
by missing out beginning and end periods that hold only misssing values.

Usage
=====

    scanFAAM.py -f <filename> [-m <missing_value>]

Where:
------

    filename   - path to a FAAM NetCDF file
    missing_value - missing value to use for variables.

"""

import os, sys, cdms, getopt

def scanFAAM(fileName=None, vars=None, nth=4, missingValuesToUse=(-9999., -32767.)):
    """
    Scans every 'nth' variable in the list of variables (or found in the
    file and gets the first and last index of the first (time) dimension
    that holds real (non-missing) values.
    """
    if type(missingValuesToUse)!=type((1,2)):
        missingValuesToUse=(missingValuesToUse,)
    startList=[]
    endList=[]
    start=None
    end=None

    if not fileName and not vars:
        raise "You must provide either a file name or a list of cdms variables."
    
    if fileName:
        f=cdms.open(fileName)
        vars=f.listvariables()

    for var in vars:
        if type(var)!=type(""):
            id=var.id
        else:
            id=var

        if id[-4:]=="FLAG" or id=="Time":
            continue

        if type(var)==type(""):
            var=f(var)

        step=1000
        while (start, end)==(None, None):
            (start, end)=findMissing(var, step, missingValuesToUse)
            step=step/2

        startList.append(start)
        endList.append(end)
        print "Start/End index: %s %s:%s" % (id, start, end)
  
    startMin=min(startList)
    endMax=max(endList)
    return (startMin, endMax)

            
def findMissing(var, step, missingValuesToUse):
    """
    Returns the (start, end) tuple for a given variable where
    they are indices of an array where missing values end and begin.
    """
    start=None
    end=None
    i0=0    
    sh=var.shape
    iend=sh[0]-1

    print var.id, step
    for miss in missingValuesToUse:
        for i in range(i0, iend, step):
            if var[i][0]==miss:
                start=i
                break
            
        for e in range(iend, i0, -step):
            if var[e][0]==miss:
                end=e
                break
            
    return (start, end)

if __name__=="__main__":

    argList=sys.argv[1:]
    args=getopt.getopt(argList, "f:m:")
    fileName=None
    missingValue=None

    for arg,value in args[0]:
        if arg=="-f":
            fileName=value
        elif arg=="-m":
            missingValue=float(value)

    scanFAAM(fileName=fileName, missingValuesToUse=(missingValue,))



#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
cdms_objs_to_na_file.py
========================

Holds the class CDMSObjectsToNAFile (sub-classing NCToNA) that converts a 
CDMS objects (variables and global atts) to one or more NASA Ames files.

"""

# Imports from python standard library
import sys
import logging

# Import from nappy package
import nappy
from nappy.na_error import na_error
import nappy.utils
import nappy.nc_interface.nc_to_na

# Import external packages (if available)
if sys.platform.find("win") > -1:
    raise na_error.NAPlatformError("Windows does not support CDMS. CDMS is required to convert to CDMS objects and NetCDF.")

try:
    import cdms2 as cdms
except:
    try:
        import cdms
    except:
        raise Exception("Could not import third-party software. Nappy requires the CDMS and Numeric packages to be installed to convert to CDMS and NetCDF.")

cdms.setAutoBounds("off")

# Define global variables
DEBUG = nappy.utils.getDebug()
default_delimiter = nappy.utils.getDefault("default_delimiter")
default_float_format = nappy.utils.getDefault("default_float_format")

# Define logger
logging.basicConfig()
log = logging.getLogger(__name__)


class CDMSObjectsToNAFile(nappy.nc_interface.nc_to_na.NCToNA):
    """
    Converts a set of CDMS Objects to one or more NASA Ames files.
    """

    def __init__(self, cdms_variables, global_attributes=[], na_items_to_override={},
                 only_return_file_names=False,
                 requested_ffi=None,
                 ):
        """
        Sets up instance variables.
        Typical usage is:
        >>>    import nappy.nc_interface.cdms_objs_to_na_file as na_maker
        >>>    c = na_maker.CDMSObjectsToNAFile(existing_var_list)
        >>>    c.convert()
        >>>    c.writeNAFiles("new_file.na", delimiter=",")
        """
        self.cdms_variables = cdms_variables
        self.global_attributes = global_attributes
        self.na_items_to_override = na_items_to_override
        self.only_return_file_names = only_return_file_names
        self.requested_ffi = requested_ffi

        self.converted = False
        self.output_message = []




#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
nc_to_na.py
=============

Holds the class NCToNA (sub-classing CDMSToNA) that converts a NetCDF file to
one or more NASA Ames files.

"""

# Imports from python standard library
import sys
import logging

# Import from nappy package
import nappy
from nappy.na_error import na_error
import nappy.utils
import nappy.utils.common_utils
import nappy.nc_interface.cdms_to_na
import nappy.nc_interface.na_content_collector

# Import external packages (if available)
if sys.platform.find("win") > -1:
    raise na_error.NAPlatformError("Windows does not support CDMS. CDMS is required to convert to CDMS objects and NetCDF.")

try:
    import cdms2 as cdms
    import numpy
except:
    try:
        import cdms
        import numpy
    except:
        raise Exception("Could not import third-party software. Nappy requires the CDMS and Numeric packages to be installed to convert to CDMS and NetCDF.")

cdms.setAutoBounds("off") 

# Define global variables
DEBUG = nappy.utils.getDebug() 
default_delimiter = nappy.utils.getDefault("default_delimiter")
default_float_format = nappy.utils.getDefault("default_float_format")
comment_override_rule = nappy.utils.getDefault("comment_override_rule")
add_column_headers = bool(nappy.utils.getDefault("add_column_headers"))

config_dict = nappy.utils.getConfigDict()
header_partitions = config_dict["header_partitions"]
hp = header_partitions

# Define global variables
permitted_overwrite_metadata = ("DATE",  "RDATE", "ANAME", "MNAME",
           "ONAME", "ORG", "SNAME", "VNAME", "SCOM", "NCOM")
items_as_lists = ["DATE", "RDATE", "ANAME", "VNAME"]

logging.basicConfig()
log = logging.getLogger(__name__)

class NCToNA(nappy.nc_interface.cdms_to_na.CDMSToNA):
    """
    Converts a NetCDF file to one or more NASA Ames files.
    """

    def __init__(self, nc_file, var_ids=None, na_items_to_override={}, 
            only_return_file_names=False, exclude_vars=[],
            requested_ffi=None,
            ):
        """
        Sets up instance variables.
        Typical usage is:
        >>>    import nappy.nc_interface.nc_to_na as nc_to_na
        >>>    c = nc_to_na.NCToNA("old_file.nc") 
        >>>    c.convert()
        >>>    c.writeNAFiles("new_file.na", delimiter=",") 

        OR:
        >>>    c = nc_to_na.NCToNA("old_file.nc") 
        >>>    file_names = c.constructNAFileNames() 
        """
        self.nc_file = nc_file

        # Now need to read CDMS file so parent class methods are compatible
        (cdms_variables, global_attributes) = self._readCDMSFile(var_ids, exclude_vars)
        nappy.nc_interface.cdms_to_na.CDMSToNA.__init__(self, cdms_variables, global_attributes=global_attributes, 
                                                        na_items_to_override=na_items_to_override, 
                                                        only_return_file_names=only_return_file_names,
                                                        requested_ffi=requested_ffi)
 

    def _readCDMSFile(self, var_ids=None, exclude_vars=[]):
        """
        Reads the file and returns all the CDMS variables in a list as well
        as the global attributes: (cdms_variable_list, global_atts_list)
        If var_ids is defined then only get those.
        """
        fin = cdms.open(self.nc_file)
        cdms_variables = []

        # Make sure var_ids is a list
        if type(var_ids) == type("string"):
            var_ids = [var_ids]

        for var_id in fin.listvariables():
            if var_ids == None or var_id in var_ids:
                if var_id not in exclude_vars:

                    # Check whether singleton variable, if so create variable
                    vm = fin[var_id]
                    var = fin(var_id)
                   
                    if hasattr(vm, "rank") and vm.rank() == 0:
                        var = cdms.createVariable(numpy.array(float(fin(var_id))), id=vm.id, attributes=vm.attributes)

                    cdms_variables.append(var)

        globals = fin.attributes.items()
        return (cdms_variables, globals) 

    def constructNAFileNames(self, na_file=None):
        """
        Works out what the file names of the output NA files will be and 
        returns a list of them.
        """
        self.convert()

        file_names = []
        # create file name if not given
        if na_file == None:
            base_name = self.nc_file
            if base_name[-3:] == ".nc":
                base_name = base_name[:-3]
            na_file = base_name + ".na"

        file_counter = 1
        # Now, create some valid file names
        for this_na_dict in self.na_dict_list:
            if len(self.na_dict_list) == 1:
                suffix = ""
            else:
                suffix = "_%s" % file_counter

            # Create file name
            name_parts = na_file.split(".")    
            new_name = (".".join(name_parts[:-1])) + suffix + "." + name_parts[-1]
            file_names.append(new_name)
            file_counter += 1
	    
        return file_names

    def writeNAFiles(self, na_file=None, delimiter=default_delimiter, annotation=False,
                     float_format=default_float_format, size_limit=None, no_header=False):
        """
        Writes the self.na_dict_list content to one or more NASA Ames files.
        Output file names are based on the self.nc_file name unless specified
        in the na_file_name argument in which case that provides the main name
        that is appended to if multiple output file names are required.

        TODO: no_header is NOT implemented.
        """
        self.convert() # just in case not already called

        # Gets a list of NA file_names that will be produced.
        file_names = self.constructNAFileNames(na_file)

        # Set up some counters: file_counter is the expected number of files.
        #      full_file_counter includes any files that have been split across multiple output NA files
        #              because size_limit exceeded.
        file_counter = 1
        full_file_counter = 1
        file_list = []

        # Get any NASA Ames dictionary values that should be overwritten with local values
        local_attributes = nappy.utils.getLocalAttributesConfigDict()
        local_na_atts = local_attributes["na_attributes"]

        # define final override list by using defaults then locally provided changes
        overriders = local_na_atts
        for (okey, ovalue) in self.na_items_to_override.items():
            overriders[okey] = ovalue

        # Now loop through writing the outputs
        for na_dict_and_var_ids in self.na_dict_list:
            file_name = file_names[file_counter - 1]
            msg = "\nWriting output NASA Ames file: %s" % file_name
            if DEBUG: log.debug(msg)
            self.output_message.append(msg)

            # Set up current na dict
            (this_na_dict, vars_to_write) = na_dict_and_var_ids

            # Override content of NASA Ames if they are permitted
            for key in overriders.keys():

                if key in permitted_overwrite_metadata:    
                    if key in items_as_lists:
                        new_item = overriders[key].split()		   
                        if key in ("DATE", "RDATE"):
                            new_item = [int(list_item) for list_item in new_item]
                    else:
                        new_item = overriders[key]

                    # Do specific overwrite for comments by inserting lines at start
                    if key in ("SCOM", "NCOM"):

                        # Use rule defined in config file in terms of where to put new comments
                        if comment_override_rule == "replace":
                            comments_list = new_item[:]
                        elif comment_override_rule in ("insert", "extend"):
                            new_comments = new_item[:]
                            existing_comments = this_na_dict.get(key, [])
                            comments_list = self._cleanWrapComments(existing_comments, new_comments, key, comment_override_rule)
                        else:
                            raise Exception("Did not recognise comment_override_rule: " + str(comment_override_rule))

                        this_na_dict[key] = comments_list
                        this_na_dict["N%sL" % key] = len(comments_list)
		    	 
                    elif not this_na_dict.has_key(key) or new_item != this_na_dict[key]:
                        this_na_dict[key] = new_item
                        msg = "Metadata overwritten in output file: '%s' is now '%s'" % (key, this_na_dict[key])
                        if DEBUG: log.debug(msg)
                        self.output_message.append(msg)

            # For certain FFIs create final Normal comments as a list of column headers before data section 
            if add_column_headers == True:
                self._updateWithColumnHeaders(this_na_dict, delimiter)
        
            # Cope with size limits if specified and FFI is 1001
            # Seems to be writing different chunks of a too long array to different na_dicts to then write to separate files.
            if size_limit is not None and (this_na_dict["FFI"] == 1001 and len(this_na_dict["V"][0]) > size_limit):
                files_written = self._writeNAFileSubsetsWithinSizeLimit(this_na_dict, file_name, delimiter=delimiter,
                                                                        float_format=float_format, size_limit=size_limit,
                                                                        annotation=annotation)
                file_list.extend(files_written)

            # If not having to split file into multiple outputs (normal condition)
            else:		
                log.info("Output NA file name: %s" % file_name)
                x = nappy.openNAFile(file_name, 'w', this_na_dict)
                x.write(delimiter=delimiter, float_format=float_format, annotation=annotation)
                x.close()
                file_list.append(file_name)

            # Report on what has been written
            msg = "\nWrote the following variables:" + "\n  " + ("\n  ".join(vars_to_write[0]))
            if DEBUG: log.debug(msg)
            self.output_message.append(msg)
	
            msg = ""
            aux_var_count = vars_to_write[1]
            if len(aux_var_count) > 0:
                msg = "\nWrote the following auxiliary variables:" + "\n  " + ("\n  ".join(aux_var_count))	
	    
            singleton_var_count = vars_to_write[2]
            if len(singleton_var_count) > 0:
                msg = "\nWrote the following Singleton variables:" + "\n  " + ("\n  ".join(singleton_var_count))

            if len(file_list) > 0:
                msg = msg + ("\n\nNASA Ames file(s) written successfully: \n%s" % "\n".join(file_list))

            full_file_counter += len(file_list)
            file_counter += 1

            if DEBUG: log.debug(msg)
            self.output_message.append(msg)
	    
        full_file_count = full_file_counter - 1
        if full_file_count == 1:
            plural = ""
        else:
            plural = "s"	      
        msg = "\n%s file%s written." % (full_file_count, plural)
    
        if DEBUG: log.debug(msg)
        self.output_message.append(msg)
        self.output_files_written = file_list

        return self.output_message

    def _writeNAFileSubsetsWithinSizeLimit(self, this_na_dict, file_name, delimiter, 
                      float_format, size_limit, annotation):
        """
        If self.size_limit is specified and FFI is 1001 we can chunk the output into 
        different files in a NASA Ames compliant way. 
        Returns list of file names of outputs written.
        """
        file_names = []
        var_list = this_na_dict["V"]
        array_length = len(var_list[0])
        nvol_info = divmod(array_length, size_limit)
        nvol = nvol_info[0]

        # create the number of volumes (files) that need to be written.
        if nvol_info[1] > 0: nvol = nvol + 1

        start = 0
        letter_count = 0
        ivol = 0

        # Loop through until full array length has been written to a set of files.
        while start < array_length:
            ivol = ivol + 1
            end = start + size_limit

            if end > array_length:
                end = array_length

            current_block = []
            # Write new V array
            for v in var_list:
                current_block.append(v[start:end])

            # Adjust X accordingly in the na dictionary, because independent variable has been reduced in size
            na_dict_copy = nappy.utils.common_utils.modifyNADictCopy(this_na_dict, current_block, 
                                                                      start, end, ivol, nvol)
            # Append a letter to the file name for writing this block to
            file_name_plus_letter = "%s-%.3d.na" % (file_name[:-3], ivol)
            file_list.append(file_name_plus_letter)

            # Write data to output file
            x = nappy.openNAFile(file_name_plus_letter, 'w', na_dict_copy)
            x.write(delimiter=delimiter, float_format=float_format, annotation=annotation)
            x.close()

            msg = "\nOutput files split on size limit: %s\nFilename used: %s" % (size_limit, file_name_plus_letter)
            if DEBUG: log.debug(msg)
            self.output_message.append(msg)
            letter_count = letter_count + 1
            start = end

            file_names.append(file_name_plus_letter) 

        return file_names


    def _updateWithColumnHeaders(self, na_dict, delimiter):
        """
        Updates the NCOM and NCOML parts of the na_dict so that 
        the last normal comments line is in fact a set of headers 
        for the data section. E.g.:

           UTs     Spd  Direc
           30446.9 305  2592
           20447.9 204  2596

        The 'delimiter' argument is used to separate out the arguments.

        This option is only compatible with a limited range of FFIs and
        only works if there are no Auxiliary variables defined.
        """
        ffi = na_dict["FFI"]
        compatible_ffis = (1001, 1010, 2110)

        if ffi not in compatible_ffis or na_dict["NAUXV"] > 0:
            log.debug("Column Headers are not written for FFIs other than: %s" % str(compatible_ffis))
            return

        if ffi in (1001, 2110):
            col_names = [na_dict["XNAME"][0]]
        elif ffi in (1010,):
            col_names = []

        col_names.extend(na_dict["VNAME"])
        col_names_line = ",".join(col_names)
        na_dict["NCOM"].append(col_names_line) 
        na_dict["NNCOML"] = len(na_dict["NCOM"])
        return  


    def _cleanWrapComments(self, existing_comments, new_comments, key, comment_override_rule):
        """
        Combines new_comments with existing_comments where comments are
        either Special or Normal. 'key' defines this being defined as either
        "SCOM" or "NCOM". 'comment_override_rule' is either "insert" (new_comments first)
        or "extend" (existing_comments first).
        Returns a new list of combined_comments.
        """
        if existing_comments == []:   return new_comments
        if new_comments == []:        return existing_comments

        # Strip start header if used
        c1 = key[0].lower()
        start_line = hp[c1 + "c_start"]
        start_used = False

        if existing_comments[0] == start_line:
            existing_comments = existing_comments[1:]
            start_used = start_line
        
        # Now check last line
        end_line = hp[c1 + "c_end"]
        end_used = False

        if existing_comments[-1] == end_line:
            existing_comments = existing_comments[:-1]
            end_used = end_line

        # Check for alternative last line in NCOM
        if end_used == False and key == "NCOM":
            end_line2 = hp["data_next"]
            if existing_comments[-1] == end_line2:
                existing_comments = existing_comments[:-1]
                end_used = end_line2 
       
        # Now put back together
        ordered_comments = [existing_comments, new_comments]
        if comment_override_rule == "insert":
            ordered_comments.reverse() 

        combined_comments = ordered_comments[0] + ordered_comments[1]
        if start_used:
            combined_comments.insert(0, start_used)
        if end_used:
            combined_comments.append(end_used)
    
        return combined_comments




#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naFile1010.py
=============

Container module for NAFile1010 class.

"""

# 08/05/04 updated by selatham for bug fixes and new write methods

# Imports from python standard library

# Imports from local package
import nappy.utils.text_parser
import nappy.na_file.na_file_1001
import nappy.utils.common_utils
wrapLine = nappy.utils.common_utils.annotateLine
wrapLines = nappy.utils.common_utils.annotateLines

class NAFile1010(nappy.na_file.na_file.NAFile):
    """
    Class to read, write and interact with NASA Ames files conforming to the
    File Format Index (FFI) 1010.
    """

    def readHeader(self):
        """
        Reads FFI-specifc header section.
        """        
        self._readCommonHeader()
        self.DX = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), self.NIV, float)
        self.XNAME = nappy.utils.text_parser.readItemsFromLines(self._readLines(self.NIV), self.NIV, str)
        self._readVariablesHeaderSection()
        self._readAuxVariablesHeaderSection()
        self._readComments()

    def writeHeader(self):								
        """											
        Writes FFI-specifc header section.					
        """   											
        self._writeCommonHeader()   						
        self.header.write(wrapLine("DX", self.annotation, self.delimiter, ("%s " * self.NIV + "\n") % tuple(self.DX)))
        self.header.write(wrapLines("XNAME", self.annotation, self.delimiter, ("%s\n") * self.NIV % tuple(self.XNAME)))
        self._writeVariablesHeaderSection()   					
        self._writeAuxVariablesHeaderSection()   				
        self._writeComments() 							
        self._fixHeaderLength()
        self.file.write(self.header.read())

    def _setupArrays(self):
        """
        Sets up FFI-specific arrays to fill with data (lists of lists).
        """        
        self.X = []
        self.V = []
        self.A = []

        for n in range(self.NV):
            self.V.append([])
        for a in range(self.NAUXV):
            self.A.append([])

    def _readData1(self, datalines, ivar_count): 
        """
        Reads first line/section of current block of data.
        """
        # Start with independent and Auxilliary vars
        (x2_and_a, rtlines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, 1 + self.NAUXV, float)
        (x, aux) = (x2_and_a[0], x2_and_a[1:])
        self.X.append(x)

        count = 0
        for a in range(self.NAUXV):
            self.A[a].append(aux[count])
            count = count + 1
        return rtlines
   
    def _readData2(self, datalines, ivar_count):
        """
        Reads second line/section (if used) of current block of data.
        """        
        # Now get the dependent variables
        (v, rtlines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, self.NV, float)              
					
        count = 0
        for n in range(self.NV):				
            self.V[n].append(v[count])							
            count = count + 1

        return rtlines

    def writeData(self):									
         """   											
         Writes the data section of the file.   					
         This method can be called directly by the user.   			
         """   											
         for m in range(len(self.X)):   						
             # Write Independent variable mark and auxiliary variables   	
             var_string = self.format % self.X[m]   
					
             for a in range(self.NAUXV):   						
                 var_string = var_string + (self.format % self.A[a][m])
	
             self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(" ,")))	
			
             # Write dependant variables   						
             var_string = ""   								
             for n in range(self.NV):							
                  var_string = var_string + (self.format % self.V[n][m])   	
	
             self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(' ,'))) 					



#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
cdms_to_na.py
=============

Holds the class CDMSToNA that converts a set of CDMS variables and global attributes.

"""

# Imports from python standard library
import sys
import logging

# Import from nappy package
import nappy
from nappy.na_error import na_error
import nappy.utils
import nappy.utils.common_utils
import cdms_utils.var_utils
import nappy.na_file.na_core
import nappy.nc_interface.na_content_collector

# Import external packages (if available)
if sys.platform.find("win") > -1:
    raise na_error.NAPlatformError("Windows does not support CDMS. CDMS is required to convert to CDMS objects and NetCDF.")
try:
    import cdms2 as cdms
    import numpy as N
except:
    try:
        import cdms
        import Numeric as N
    except:
        raise Exception("Could not import third-party software. Nappy requires the CDMS and Numeric packages to be installed to convert to CDMS and NetCDF.")

cdms.setAutoBounds("off") 

# Define global variables
var_limit = 5000 # surely never going to get this many vars in a file!

DEBUG = nappy.utils.getDebug() 

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

class CDMSToNA:
    """
    Converts CDMS objects to NASA Ames file dictionaries.
    """

    def __init__(self, cdms_variables, global_attributes=[], na_items_to_override={}, 
                 only_return_file_names=False, requested_ffi=None,
                 ):
        """
        Sets up instance variables.      
        """
        self.cdms_variables = cdms_variables
        self.global_attributes = global_attributes
        self.na_items_to_override = na_items_to_override
        self.only_return_file_names = only_return_file_names
        self.requested_ffi = requested_ffi

        self.converted = False
        self.output_message = []
    
    def convert(self):
        """
        Reads the CDMS objects and convert to a set of dictionaries that
        provide the structure for a NA File object.
        Returns [(na_dict, var_ids), (na_dict, var_ids), ....]
        All these na_dict dictionaries can be readily written to a NA File object.

        Note that NASA Ames is not as flexible as NetCDF so you cannot just send any 
        set of variables to write to a NASA Ames file. Essentially there is one
        multi-dimensional structure and all variables must be defined against it.

        Otherwise variables must be auxiliary variables within that structure (i.e. only
        defined once per the least changing dimension.
        """
        if self.converted == True:
            return self.na_dict_list
        
        # Convert any singleton variables to CDMS variables
        variables = self._convertSingletonVars(self.cdms_variables)

        # Re-order variables if they have the attribute "nasa_ames_var_number" which means they came from a NASA Ames file originally
        variables = self._reorderVars(variables)

        # Make first call to collector class that creates NA dict from CDMS variables and global atts list 
        collector = nappy.nc_interface.na_content_collector.NAContentCollector(variables, 
                                        self.global_attributes, requested_ffi=self.requested_ffi,
                                        )
        collector.collectNAContent()

        # Return if no files returned
        if collector.found_na == False:
            msg = "\nNo files created after variables parsed."
            if DEBUG: log.debug(msg)
            self.output_message.append(msg)
            return

        # NOTE: collector has attributes: na_dict, var_ids, unused_vars

        # Set up a list to collect multiple calls to content collector
        na_dict_list = []
        na_dict_list.append((collector.na_dict, collector.var_ids))

        # If there are variables that were not captured (i.e. unused) by NAContentCollector then loop through these
        # in attempt to convert all to a set of na_dicts
        log.debug("\nUnused_vars: %s" % collector.unused_vars)
        while len(collector.unused_vars) > 0:
            collector = nappy.nc_interface.na_content_collector.NAContentCollector(collector.unused_vars, 
                                        self.global_attributes, requested_ffi=self.requested_ffi,
                                        )
            collector.collectNAContent()           
            self.output_message += collector.output_message

            # Append to list if more variables were captured
            if collector.found_na == True:  
                na_dict_list.append((collector.na_dict, collector.var_ids))

        self.na_dict_list = na_dict_list
        self.converted = True

        return self.na_dict_list

    def _convertSingletonVars(self, variables):
        """
        Loops through variables to convert singleton variables (i.e. Masked Arrays/Numeric Arrays) 
        to proper CDMS variables. Then code won't break when asking for rank attribute later.
        Returns a list of CDMS variable objects
        """
        vars = []

        for variable in variables:
            var_obj = variable

            # If singleton variable then convert into proper CDMS variables so code doesn't break later
            if not hasattr(var_obj, "rank") or var_obj.rank() == 0:
              
                var_metadata = var_obj.attributes       
                var_obj = cdms.createVariable(N.array(var_obj), 
                                   id = cdms_utils.var_utils.getBestName(var_metadata).replace(" ", "_"), 
                                   attributes=var_metadata)
                var_obj.value = float(var_obj._data)
                
            vars.append(var_obj)

        return vars

    def _reorderVars(self, variables):
        """
        Returns a reordered list of variables. Any that have the attribute 
        "nasa_ames_var_number" get ordered first in the list (according to numbering).
        """
        # Set up a long list (longer than number of vars)
        if len(variables) > var_limit:
            raise Exception("Can only handle converting less than " + `var_limit` + " variables in any batch.")

        # Collect up those that are ordered and unordered
        ordered_vars = [None] * var_limit
        unordered_vars = []

        for var in variables:
            var_metadata = var.attributes
            if hasattr(var_metadata, "nasa_ames_var_number"):
                num = var_metadata.nasa_ames_var_number
                ordered_vars[num] = var
            else:
                unordered_vars.append(var)
    
        vars = []

        # Clear any None values in ordered_vars and place in final vars list
        for var in ordered_vars + unordered_vars:
            # Test for Real var types
            if type(var) != type(None): 
                vars.append(var)
	     
        return vars




#!/usr/bin/env python
#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""

nc2csv.py
=========

Converts a NetCDF file into one or more NASA Ames file.

Usage
=====

    nc2csv.py [-v <var_list>] [--ffi=<ffi>] [-f <float_format>] 
             [-l <limit_ffi_1001_rows>] 
             [-e <exclude_vars>] [--overwrite-metadata=<key1>,<value1>[,<key2>,<value2>[...]]] 
             [--names-only] [--no-header] [--annotated]
             -i <nc_file> [-o <csv_file>] 
Where
-----

    <nc_file> 			- name of input file (NetCDF).
    <csv_file>	 		- name of output file (NASA Ames or CSV) - will be used as base name if multiple files.
    <var_list>           	- a comma-separated list of variables (i.e. var ids) to include in the output file(s).
    <ffi>			- NASA Ames File Format Index (FFI) to write to (normally automatic).  
    <float_format>          	- a python formatting string such as %s, %g or %5.2f
    <limit_ffi_1001_rows> 	- if format FFI is 1001 then chop files up into <limitFFI1001Rows> rows of data.  
    <exclude_vars>          	- a comma-separated list of variables (i.e. var ids) to exclude in the output file(s).
    <key1>,<value1>[,<key2>,<value2>[...]] - list of comma-separated key,value pairs to overwrite in output files:
								* Typically the keys are in: 
                                   * "DATE", "RDATE", "ANAME", "MNAME","ONAME", "ORG", "SNAME", "VNAME".
    --names-only		- only display a list of file names that would be written (i.e. don't convert actual files).
    --no-header			- Do not write NASA Ames header
    --annotated			- add annotation column in first column
    
"""

# Imports from python standard library
import sys
import getopt

# Import from nappy package
import nc2na

# Uses nc2na(...) directly



if __name__ == "__main__":

    args = sys.argv[1:]
    if "-d" in args:
        exitNicely("nc2csv.py does not accept the -d argument as the delimiter is fixed as a comma, try using nc2na.py instead.")

    args.extend(["-d", ","])
    nc2na.nc2na(args)



import na_to_nc
import nc_to_na
import os

in_dir = "../../data_files"
out_dir = "../../test_outputs"
in_dir = out_dir

for ffi in (1001, 2010, 3010, 4010):

    infile = os.path.join(in_dir, "%s.nc" % ffi)
    outfile = os.path.join(out_dir, "%s_from_nc.na" % ffi)

    print "Reading:", infile
    x = nc_to_na.NCToNA(infile)

    print "Writing:", outfile
    x.writeNAFiles(outfile)



#    Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#    This software may be distributed under the terms of the
#    Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
common_utils.py
===============

Functions and classes commonly used in nappy.

"""

# Standard library imports
import logging

# Imports from local package
import parse_config
import text_parser as text_parser

logging.basicConfig()
log = logging.getLogger(__name__)


def getNAFileClass(ffi):
    """
    Returns class for an FFI.
    """
    mod = "nappy.na_file.na_file_" + `ffi`
    cls = "NAFile" + `ffi`
    exec "import %s" % mod
    return eval("%s.%s" % (mod, cls))
   

def readFFI(filename):
    """
    Function to read the top line of a NASA Ames file to extract
    the File Format Index (FFI) and return it as an integer.
    """
    fin = open(filename)
    topline = fin.readline()
    fin.close()

    ffi = text_parser.readItemsFromLine(topline, 2, int)[-1]
    return ffi


def chooseFFI(na_dict):
    """
    Function to choose the appropriate FFI based on the contents of the
    'na_dict' dictionary object that holds NASA Ames internal variables.
    """
    d = na_dict

    if d["NIV"] > 4:      # More than 4 independent variables not allowed
        raise Exception("NASA Ames cannot write more than 4 independent variables.")

    elif d["NIV"] == 4:     # 4 independent variables
        return 4010

    elif d["NIV"] == 3:     # 3 independent variables
        return 3010

    elif d["NIV"] == 2:    # 2 independent variables
        if type(d["X"][0][0]) == type("string"):
            # 2160 - the  independent unbounded variable is a character string
            return 2160
        elif type(d["X"][0][1]) == type([1,2]) and len(d["X"][0][1]) > 1:
            # 2110 - one independent variable changes length and the values are specified
            return 2110
        elif type(d["X"][0][1]) == type([1,2]) and len(d["X"][0][1]) == 1:
            # 2310 - one indepenent variable changes length but only the first value is specifically stated
            return 2310
        else:            
            # 2010 - Straightforward 2-D variables
            return 2010

    elif d["NIV"] == 1:   # 1 independent variable 
        if not d.has_key("NAUXV"):
            # 1001 - No auxiliary variables
            return 1001
        elif d.has_key("NVPM"):
            # 1020 - Implied values for independent variable
            return 1020
        else:
            # 1010 - Auxiliary variables included
            return 1010
    else:
        raise Exception("Could not resolve the dictionary object to create a suitable NASA Ames File Format Index (FFI). Please modify the contents and try again.")


def getFileNameWithNewExtension(input_file, format):
    """
    Takes an input_file name and applies new extension to it by:
    (i) replacing initial extension if there is one, OR
    (ii) just appending new extension.
    """
    base_name = input_file
    last_four = base_name[-4:]
    found = last_four.find(".")
    if found > -1:
        idx = len(base_name) + found
        base_name = base_name[:idx]
    return base_name + "." + format


def modifyNADictCopy(indict, v_new, start, end, ivol, nvol):
    """
    Returns a copy of a dictionary with some modifications.
    """
    newDict = {}
    for key,value in indict.items(): 
        if key == "X":
            newlist = indict["X"][start:end]
            newDict["X"] = newlist
        elif key == "V":
            newDict["V"] = v_new
        elif key == "IVOL":
            newDict["IVOL"] = ivol
        elif key == "NVOL":
            newDict["NVOL"] = nvol
        else:
            newDict[key] = value
    return newDict


def getVersion():
    """
    Gets config dict for version.
    """
    version = parse_config.getConfigDict()["main"]["version"]
    return version


def getDebug():
    """
    Returns true or false for DEBUG status.
    """
    DEBUG = parse_config.getConfigDict()["main"]["DEBUG"]
    return eval(DEBUG)


def getDefault(item):
    """
    Returns value of item from 'main' config file section.
    """
    value =  parse_config.getConfigDict()["main"][item]
    return value


def makeDictFromCommaSepString(s):
    """
    Reads in comma-separated list and converts to dictionary of successive
    keyword,value pairs.
    """
    if s.count(",") % 2 == 0:
        raise Exception("Must provide even number of items in argument of commas-separated pairs of values: " + s)

    d = {}
    items = s.split(",")
    while len(items) > 0:
        d[items[0]] = items[1]
        items = items[2:] 
    return d


def makeListFromCommaSepString(s):
    """
    Reads in comma-separated list and converts to list of successive
    keyword,value pairs.
    """
    if s.count(",") % 2 == 0:
        raise Exception("Must provide even number of items in argument of commas-separated pairs of values: " + s)

    l = []
    items = s.split(",")
    while len(items) > 0:
        l.append((items[0], items[1]))
        items = items[2:]
    return l


def getAnnotation(item, annotation, delimiter=None, count=None):
    """
    Returns the annotation string for a given NASA Ames item.
    """
    if delimiter == None:
        delimiter = getDefault("default_delimiter") 
    dict = parse_config.getAnnotationsConfigDict()

    if count == None:
        count_string = ""
    else:
        count_string = " %s" % count
    
    if annotation:
        return "%s%s%s" % (dict[item], count_string, delimiter)
    else:
        return ""


def annotateLines(item_name, annotate, delimiter, lines):
    """
    Takes item_name to look up, delimiter and item to render and returns full line.
    """

    split_lines = lines.splitlines(1)
    output = ""

    count = 1
    for line in split_lines:
        output = output + annotateLine(item_name, annotate, delimiter, line, count)
        count += 1

    return output

    
def annotateLine(item_name, annotate, delimiter, line, count=None):
    """
    Takes item_name to look up, delimiter and item to render and returns full line.
    """

    if annotate:
        annotation = getAnnotation(item_name, annotate, delimiter=delimiter, count=count)
        line = "%s%s" % (annotation, line)
        return line
    else:
        return line


def stripQuotes(s):
    "Strips extra quotes"
    if type(s) != type("string"): s  = str(s)
    if s[0] in ("'", '"'): s = s[1:]
    if s[-1] in ("'", '"'): s = s[:-1]
    return s



#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naFile2160.py
=============

Container module for NAFile2160 class.

"""

# Imports from python standard library

# Imports from local package
import nappy.utils.text_parser
import nappy.na_file.na_file_2110
import nappy.utils.common_utils
wrapLine = nappy.utils.common_utils.annotateLine
wrapLines = nappy.utils.common_utils.annotateLines

class NAFile2160(nappy.na_file.na_file_2110.NAFile2110):
    """
    Class to read, write and interact with NASA Ames files conforming to the
    File Format Index (FFI) 2160.
    """

    def readHeader(self):
        """
        Reads FFI-specifc header section.
        """
        self._normalized_X = False
        self._readCommonHeader()
        self.DX = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), 1, float)
        self.LENX = nappy.utils.text_parser.readItemFromLine(self.file.readline(), float)
        self.XNAME = nappy.utils.text_parser.readItemsFromLines(self._readLines(self.NIV), self.NIV, str)
        self.XNAME.reverse()  # Reverse because C-type array is least-changing first
        self._readVariablesHeaderSection()
        self._readCharAuxVariablesHeaderSection()
        self._readComments()

    def writeHeader(self):
        """
        Writes FFI-specific header section.
        """
        self._writeCommonHeader()
        DX = self.DX
        DX.reverse()
        self.header.write(wrapLine("DX", self.annotation, self.delimiter, "%s\n" % tuple(DX)))
        self.header.write(wrapLine("LENX", self.annotation, self.delimiter, "%s\n" % self.LENX))
        XNAME = self.XNAME
        XNAME.reverse()
        self.header.write(wrapLines("XNAME", self.annotation, self.delimiter, "%s\n" * self.NIV % tuple(XNAME)))
        self._writeVariablesHeaderSection()
        self._writeAuxVariablesHeaderSection()
        self._writeComments()
        self._fixHeaderLength()
        self.file.write(self.header.read())

    def _writeAuxVariablesHeaderSection(self):
        """
        Writes the auxiliary variables section of the header for FFI 2160.
        Assumes we are at the right point in the file.
        """
        self.header.write(wrapLine("NAUXV", self.annotation, self.delimiter, "%d\n" % self.NAUXV))
        self.header.write(wrapLine("NAUXC", self.annotation, self.delimiter, "%d\n" % self.NAUXC))
        if self.NAUXV > 0:
            line = (("%s" + self.delimiter) * (self.NAUXV - self.NAUXC - 1) + "%s\n")  % tuple(self.ASCAL)
            self.header.write(wrapLine("ASCAL", self.annotation, self.delimiter, line))
            line = (("%s" + self.delimiter) * (self.NAUXV - self.NAUXC - 1) + "%s\n")  % tuple(self.AMISS[0:(self.NAUXV - self.NAUXC)])
            self.header.write(wrapLine("AMISS", self.annotation, self.delimiter, line))
            line = (("%s" + self.delimiter) * (self.NAUXC - 1) + "%s\n") % tuple(self.LENA[(self.NAUXV - self.NAUXC):])
            self.header.write(wrapLine("LENA", self.annotation, self.delimiter, line))
            line = ("%s\n" * self.NAUXC) % tuple(self.AMISS[(self.NAUXV - self.NAUXC):])
            self.header.write(wrapLines("AMISS", self.annotation, self.delimiter, line))
            line = "%s\n" * self.NAUXV % tuple(self.ANAME)
            self.header.write(wrapLines("ANAME", self.annotation, self.delimiter, line))

    def _setupArrays(self):
        """
        Sets up FFI-specific arrays to fill with data (lists of lists).
        """
        self.V = []
        self.A = []
        self.X = []
        self.NX = []
        
        for n in range(self.NV):
            self.V.append([])
        for i in range(self.NAUXV):
            self.A.append([])

    def _readData1(self, datalines, ivar_count):
        """
        Reads first line/section of current block of data.
        """      
        # Start with independent and Auxilliary vars
        # Get character string independent variable
        (x1, datalines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, 1, str)
	self.X.append([])
        self.X[ivar_count].append(x1[0])
        # Set up list to take second changing independent variable
        self.X[ivar_count].append([])  

        # Get NX and Non-character AUX vars
        (aux, datalines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, (self.NAUXV - self.NAUXC), float)
        self.NX.append(int(aux[0]))

        count = 0
        for a in range(self.NAUXV - self.NAUXC):
            self.A[a].append(aux[count])
            count = count + 1

        # Get character AUX vars
        (auxc) = nappy.utils.text_parser.readItemsFromLines(datalines[:self.NAUXC], self.NAUXC, str)
        rtlines = datalines[self.NAUXC:]
        count = 0
        for a in range(self.NAUXC):
            self.A[(self.NAUXV - self.NAUXC) + a].append(auxc[count])
            count = count + 1

        return rtlines
    
    def writeData(self):
        """
        Writes the data section of the file.
        This method can be called directly by the user.
        """
        # Set up unbounded IV loop
        for m in range(len(self.X)):

            # Write Independent variable mark and auxiliary variables
            var_string = "%s" % self.X[m][0]
            self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(" ,")))
            var_string = ""

            for a in range(self.NAUXV - self.NAUXC):
                var_string = var_string + (self.format % self.A[a][m])

            self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(" ,")))

            for a in range(self.NAUXC):
                var_string = (("%s" + self.delimiter) % self.A[(self.NAUXV - self.NAUXC) + a][m])
                self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(" ,")))

            # Write second independant variable and dependant variables
            for p in range(self.NX[m]):
                var_string = self.format % self.X[m][1][p]

                for n in range(self.NV):
                    var_string = var_string + (self.format % self.V[n][m][p])

                self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(" ,")))



#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
airCraftData.py
===============

Container module for AircraftData class that deals with
flight data which has unusual time axes and flags in separate 
variables. This class also does sub-sampling by averaging to 
1Hz frequency.


"""

# Imports from python standard library
import os,sys
import MA, Numeric, cdms, MV, cdtime

# Imports from local package


class AircraftData:
    """
    AircraftData class that deals with
    flight data which has unusual time axes and flags in separate
    variables. This class also does sub-sampling by averaging to
    1Hz frequency.
    """

    def __init__(self, var, timeVar, flagVar=None, args=["av"]):
        """
        Parses args and calls required methods.
        """
        if args[0]=="av":
            samplingMethod="averaging"
        elif args[0]=="ss":
            samplingMethod="selection"

        flagsToUse=[0,1]
        samplingRate=1
        if len(args)>1:
            samplingRate=int(args[1])

            if len(args)>2:
                flagsToUse=[int(flag[0]) for flag in args[2:]]

        if samplingMethod=="averaging":
            self.subSampledVar=self._subSampleByAveraging(var, timeVar, flagVar,
                    samplingRate, flagsToUse)
        elif samplingMethod=="selection":
            self.subSampledVar=self._subSampleBySelection(var, timeVar, flagVar,
                    samplingRate, flagsToUse)


    def _subSampleBySelection(self, var, timeVar, flagVar, samplingRate, flagsToUse):
        """
        Returns a new variable which is 'var' sub-sampled by selecting the value
        at the given samplingRate with data selected according to flagVar
        including the flag values  specified in flagsToUse (defaults are 0 and 1).
        """
        if samplingRate!=1:
            raise "Sampling rates of values other than 1Hz are not yet supported."
        maskedArray=self._getMaskedArray(var[:,0](squeeze=1), flagVar[:,0](squeeze=1), flagsToUse)
        # Now re-construct variable axes etc
        newTimeAxis=self._flatten2DTimeAxis(timeVar, samplingRate)
        newVar=self._recreateVariable(var, maskedArray, newTimeAxis, flagVar, max(flagsToUse), missingValue=maskedArray.fill_value(), sampleBy="selection")
        return newVar

        
    def _subSampleByAveraging(self, var, timeVar, flagVar, samplingRate, flagsToUse):
        """
        Returns a new variable which is 'var' sub-sampled by averaging
        at the given samplingRate with data selected according to flagVar
        including the flag values specified in flagsToUse (defaults are 0 and 1).
        """
        maskedArray=self._getMaskedArray(var, flagVar, flagsToUse)
        shape=var.shape
        if shape[1]==1:
            newNumArray=MV.ravel(maskedArray)
            newArrayMask=MV.ravel(maskedArray.mask())
            newArray=MA.masked_array(newNumArray, mask=newArrayMask, fill_value=maskedArray.fill_value())
        else:
            newArray=Numeric.zeros(shape[0], 'f')
            for t0 in range(shape[0]):
                # Set as missing if less than half are valid                 
                t1Array=maskedArray[t0]

                if samplingRate==1:
                    # If half or more are good values then calculate the mean
                    if t1Array.count()>=(shape[1]/2.):
                        newArray[t0]=MA.average(t1Array)
                    # otherwise set as missing value
                    else:
                        newArray[t0]=maskedArray.fill_value()
        
                else:
                    raise "Averaging for non 1Hz sampling rates not yet supported!"

        # Now re-construct variable axes etc
        newTimeAxis=self._flatten2DTimeAxis(timeVar, samplingRate)
        newVar=self._recreateVariable(var, newArray, newTimeAxis, flagVar, max(flagsToUse), missingValue=maskedArray.fill_value(), sampleBy="averaging") 
        return newVar


    def _flatten2DTimeAxis(self, timevar, samplingRate):
        """
        Returns a flattened 2D time axis.
        """
        if samplingRate!=1:
            raise "Cannot yet deal with sub-sampling to non 1Hz sampling rates!"

        timevalues=timevar._data
        timeunit=timevar.units
        newTimeValues=[]
        rate=samplingRate

        for t in timevalues:
            for i in range(rate):
                tvalue=t+((1./rate)*i)
                newTimeValues.append(tvalue)

        newTimeAx=cdms.createAxis(newTimeValues)
        newTimeAx.units=timeunit
        newTimeAx.designateTime()
        newTimeAx.id=newTimeAx.long_name=newTimeAx.standard_name="time"
        return newTimeAx


    def _recreateVariable(self, var, newArray, timeAxis, flagVar, flagMax, missingValue, sampleBy="averaging"):
        """
        Rebuilds a variable using new array and new time axis.
        """
        atts=var.attributes
        newAtts={}
        for att,value in atts.items():
            if type(value) in (type((1,1)), type([1,2]), type(Numeric.array([1.]))) and len(value)==1:
                value=value[0]
            if type(value) in (type(1), type(1.), type(long(1))):
                newAtts[att]=value
            elif type(value)==type(""):
                newAtts[att]=value.strip()

        # Now create the variable
        missing=missingValue
        data=newArray
        newvar=cdms.createVariable(data, id=var.id, fill_value=missing, axes=[timeAxis], attributes=newAtts)
        newvar.frequency=1
        if sampleBy=="averaging":
            newvar.comment="Data averaged to give a pseudo sampling frequency of 1 Hz. "
            if flagVar:
                newvar.comment=newvar.comment+"Values were included where the flag variable was less than %s and where both the flag value and actual data value were present (i.e. not missing values)." % (flagMax+1)
                newvar.comment=newvar.comment+" Average data values were set to missing where more than half the values in any one second period were already missing. Missing value was also used where more than half the flags were either missing value or values of 2 or greater."
            else:
                newvar.comment=newvar.comment+"The flag variable was not used in creating this variable."
                newvar.comment=newvar.comment+" Average data values were set to missing where more than half the values in any one second period were already missing."
        elif sampleBy=="selection":
            newvar.comment="Data sub-sampled to give a pseudo sampling frequency of 1 Hz by selecting only the per-second frequency values. "
            if flagVar:
                newvar.comment=newvar.comment+" Values were included where the flag variable was less than 2 and where both the flag value and actual data value were present (i.e. not missing values)."
            else:
                newvar.comment=newvar.comment+" The flag variable was not used in creating this variable." 
        return newvar
  

    def _getMaskedArray(self, var, flagVar=None, flagValues=(0,1), missingValuesToTest=(-9999., -32767.)):
        """
        Returns a masked array that has the values only present where the flag
        values are acceptable and the data itself is not missing.
        """
        if flagVar:
            flagFillValue=-1
            flagGTValue=max(flagValues)

            flagMaskWhereUnacceptable=MA.masked_greater(flagVar, flagGTValue).mask()
            flagMaskWhereMissing=MA.masked_equal(flagVar, flagFillValue).mask()
            flagMask=flagMaskWhereUnacceptable+flagMaskWhereMissing

        for fv in missingValuesToTest:
            if fv in MV.ravel(var):
                print "Setting missing value for '%s' as: %s" % (var.id, fv)
                varFillValue=fv
        else:
            varFillValue=missingValuesToTest[0]

        if flagVar:
            varMask=MA.masked_array(var, mask=MA.equal(var, varFillValue), fill_value=varFillValue).mask()
            fullmask=MA.bitwise_or(varMask, flagMask)
            maskedArray=MA.masked_array(var, mask=fullmask, fill_value=varFillValue)
        else:
            maskedArray=MA.masked_array(var, mask=MA.equal(var, varFillValue), fill_value=varFillValue)

        #maskedArray=cdms.createVariable(maskedArray, id=var.id, fill_value=varFillValue)
        #maskedArray.missing_value=varFillValue
        return maskedArray 




#!/usr/bin/env python
#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""

nc2na.py
========

Converts a NetCDF file into one or more NASA Ames file.

Usage
=====

    nc2na.py [-v <var_list>] [--ffi=<ffi>] [-f <float_format>] 
             [-d <delimiter>] [-l <limit_ffi_1001_rows>] 
             [-e <exclude_vars>] [--overwrite-metadata=<key1>,<value1>[,<key2>,<value2>[...]]] 
             [--names-only] [--no-header] [--annotated]
             -i <nc_file> [-o <na_file>] 
Where
-----

    <nc_file> 			- name of input file (NetCDF).
    <na_file>	 		- name of output file (NASA Ames or CSV) - will be used as base name if multiple files.
    <var_list>           	- a comma-separated list of variables (i.e. var ids) to include in the output file(s).
    <ffi>			- NASA Ames File Format Index (FFI) to write to (normally automatic).  
    <float_format>          	- a python formatting string such as %s, %g or %5.2f
    <delimiter>	 		- the delimiter you wish to use between data items in the output file such as "   " or "\t"
    <limit_ffi_1001_rows> 	- if format FFI is 1001 then chop files up into <limitFFI1001Rows> rows of data.  
    <exclude_vars>          	- a comma-separated list of variables (i.e. var ids) to exclude in the output file(s).
    <key1>,<value1>[,<key2>,<value2>[...]] - list of comma-separated key,value pairs to overwrite in output files:
								* Typically the keys are in: 
                                   * "DATE", "RDATE", "ANAME", "MNAME","ONAME", "ORG", "SNAME", "VNAME".
    --names-only		- only display a list of file names that would be written (i.e. don't convert actual files).
    --no-header			- Do not write NASA Ames header
    --annotated			- add annotation column in first column
    
"""

# Imports from python standard library
import sys
import getopt

# Import from nappy package
import nappy
from nappy.utils.common_utils import makeListFromCommaSepString, makeDictFromCommaSepString


def exitNicely(msg=""):
    "Exits nicely!"
    print __doc__
    if msg != "": print "ERROR:", msg
    sys.exit()


def parseArgs(args):
    """
    Parses arguments and returns dictionary.
    """
    arg_dict = {}
    a = arg_dict

    # Set up defaults
    a["nc_file"] = None
    a["var_ids"] = None
    a["requested_ffi"] = None
    a["float_format"] = nappy.default_float_format
    a["delimiter"] = nappy.default_delimiter
    a["size_limit"] = None
    a["exclude_vars"] = []
    a["na_items_to_override"] = {}
    a["only_return_file_names"] = False
    a["na_file"] = None
    a["no_header"] = False
    a["annotation"] = False

    try:
        (arg_list, dummy) = getopt.getopt(args, "i:o:v:f:d:l:e:", 
                              ["ffi=", "overwrite-metadata=", "names-only",
                               "no-header", "annotated"])
    except getopt.GetoptError, e:
        exitNicely(str(e))

    for arg, value in arg_list:
        if arg == "-i":
            a["nc_file"] = value
        elif arg == "-o":
            a["na_file"] = value
        elif arg == "-v":
            a["var_ids"] = value.split(",")
        elif arg == "--ffi":
            a["requested_ffi"] = int(value)
        elif arg == "-f":
            a["float_format"] = value
        elif arg == "-d":
            a["delimiter"] = value
        elif arg == "--limit_ffi_1001_rows":
            a["size_limit"] = long(value)
        elif arg == "-e":
            a["exclude_vars"] = value.split(",")
        elif arg == "--overwrite-metadata":
            a["na_items_to_override"] = makeDictFromCommaSepString(value)
        elif arg == "--names-only":
            a["only_return_file_names"] = True
        elif arg == "--no-header":
            a["no_header"] = True
        elif arg == "--annotated":
            a["annotation"] = True
        else:
            exitNicely("Argument '" + arg + "' not recognised!")

    if not a["nc_file"]:
        exitNicely("Please provide argument '-i <nc_file>'")

    return a


def nc2na(args=None):
    """
    Controller for conversion of NetCDF file to NASA Ames files.
    """

    if args is None:
        args = sys.argv[1:]

    arg_dict = parseArgs(args)
    nc_file = arg_dict["nc_file"]
    del arg_dict["nc_file"]
    na_files = apply(nappy.convertNCToNA, [nc_file], arg_dict)

    # If user only wants files then only give them that
    if arg_dict["only_return_file_names"] == True:
        print "\nExpected file names would be:"
        for naf in na_files:
            print "    ", naf

    else:
        print "\nSuccessfully wrote: " 
        for naf in na_files:
            print "    ", naf


if __name__ == "__main__":

    args = sys.argv[1:]
    nc2na(args)



#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naFile1020.py
=============

Container module for NAFile1020 class.

"""

# Imports from python standard library


# Imports from local package
import nappy.utils.text_parser
import nappy.na_file.na_file_1010
import nappy.utils.common_utils
wrapLine = nappy.utils.common_utils.annotateLine
wrapLines = nappy.utils.common_utils.annotateLines

class NAFile1020(nappy.na_file.na_file_1010.NAFile1010):
    """
    Class to read, write and interact with NASA Ames files conforming to the
    File Format Index (FFI) 1020.
    """

    def readHeader(self):
        """
        Reads FFI-specifc header section.
        """    
        self._normalized_X = False
        self._readCommonHeader()
        self.DX = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), self.NIV, float)
        if self.DX == 0:
            raise "DX found to be zero (0). Not allowed for FFI 1020."

        self.NVPM = nappy.utils.text_parser.readItemFromLine(self.file.readline(), int)
        self.XNAME = nappy.utils.text_parser.readItemsFromLines(self._readLines(self.NIV), self.NIV, str)
        self._readVariablesHeaderSection()
        self._readAuxVariablesHeaderSection()
        self._readComments()
  
    def writeHeader(self):										
        """													
        Writes FFI-specifc header section.						
        """   												
        self._writeCommonHeader()   								
        self.header.write(wrapLine("DX", self.annotation, self.delimiter, ("%s " * self.NIV + "\n") % tuple(self.DX)))
        self.header.write(wrapLine("NVPM", self.annotation, self.delimiter, ("%s\n") %self.NVPM))							
        self.header.write(wrapLine("XNAME", self.annotation, self.delimiter, ("%s\n") * self.NIV % tuple(self.XNAME)))
        self._writeVariablesHeaderSection()   						
        self._writeAuxVariablesHeaderSection()   						
        self._writeComments() 								
        self._fixHeaderLength()
        self.file.write(self.header.read())

    def _readData2(self, datalines, ivar_count):
        """
        Reads second line/section (if used) of current block of data.
        """
        # Now get the dependent variables
        (v, rtlines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, self.NV * self.NVPM, float)              
        count = 0
        for n in range(self.NV):
            for i in range(self.NVPM):   # Number of steps where independent variable is implied
		self.V[n].append(v[count])						
		count = count + 1										
	return rtlines

    def _normalizeIndVars(self):
        """
        Normalizes the values in the unbounded independent variable for FFIs
        that store an abbreviated version of this axis.
        """
        if self._normalized_X == True: return
        newX = []
        for x in self.X[0]:
            for i in range(self.NVPM):
                newX.append(x + (i * self.DX))

        self.X[0] = newX
        self._normalized_X = True

    def writeData(self):										
         """   													
         Writes the data section of the file.   							
         This method can be called directly by the user.   					
         """   													
         for m in range(len(self.X)):   								
            # Write Independent variable mark and auxiliary variables   			
            var_string = self.format % self.X[m]   	
						
            for a in range(self.NAUXV):   								
                var_string = var_string + (self.format % self.A[a][m])   
			
            self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(" ,")))
				
            # Write dependant variables																		
            count = 0											
            for n in range(self.NV):									
                var_string = ""	
								        
                for p in range(self.NVPM):								
                    var_ind = (m * self.NVPM) + p 							
                    var_string = var_string + (self.format % self.V[n][var_ind])   		
                    count = count + 1	
								
                self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" %var_string.rstrip(' ,')))




"""
parse_config.py
===================

Parses config file for nappy.

"""

# Standard library imports
import ConfigParser
import os


# Global variables

# Defaults are always loaded from <nappy-egg>/nappy/config/nappy.ini
_here = os.path.dirname(__file__)
base_dir = os.path.join(_here, '../config')

config_file = os.path.join(base_dir, "nappy.ini")
config_dict = None
annotations_config_dict = None
attributes_config_dict = None

class MyCasePreservingConfigParser(ConfigParser.ConfigParser):
    optionxform = str

def makeConfigDict(cf=config_file):
    """
    Parses config file and returns dictionary of sub-dictionaries
    for each section holding Keyword-Value pairs.
    """
    d = {}
    conf = MyCasePreservingConfigParser()
    conf.read(cf)

    # get all sections and content 
    for section in conf.sections():
        d[section] = {}
        for item in conf.options(section):
            value = conf.get(section, item)
            if value.find("__space__") > -1:
                value = value.replace("__space__", " ")

            if item.find("&") > -1:
                item = tuple(item.split("&"))
            if value.find("&") > -1:
                value = tuple(value.split("&"))
            d[section][item] = value

    return d


def getConfigDict(cf=config_file):
    "Checks if already made and only makes if required."
    global config_dict
    if config_dict == None:
        config_dict = makeConfigDict(cf)
    return config_dict


def makeAnnotationsConfigDict(af):
    """
    Parses annotations config file and returns dictionary of annotations.
    """
    ad = {}
    conf = MyCasePreservingConfigParser()
    conf.read(af)

    # Load up dict
    for item in conf.options("annotations"):
        value = conf.get("annotations", item)
        ad[item] = value

    return ad


def getAnnotationsConfigDict():
    "Checks if already made and only makes if required."
    config_dict = getConfigDict()

    annotations_config_file = os.environ.get("NAPPY_ANNOTATIONS", None) or \
                              os.path.join(base_dir, config_dict["main"]["annotations_file"])

    global annotations_config_dict

    if annotations_config_dict == None:
        annotations_config_dict = makeAnnotationsConfigDict(annotations_config_file)

    return annotations_config_dict


def makeLocalAttributesConfigDict(laf):
    """
    Parses local attributes config file and returns dictionary.
    """
    lad = {}
    conf = MyCasePreservingConfigParser()
    conf.read(laf)

    # Load up dict
    for sect in ("nc_attributes", "na_attributes"):
        lad[sect] = {}
        for item in conf.options(sect):
            value = conf.get(sect, item)
            lad[sect][item] = value

    return lad


def getLocalAttributesConfigDict():
    "Checks if already made and only makes if required."
    config_dict = getConfigDict()

    local_attributes_config_file = os.environ.get("NAPPY_LOCAL_ATTRIBUTES", None) or \
                                   os.path.join(base_dir, config_dict["main"]["local_attributes_file"])
    
    global attributes_config_dict

    if attributes_config_dict == None:
        attributes_config_dict = makeLocalAttributesConfigDict(local_attributes_config_file)
    return attributes_config_dict



if __name__=="__main__":

    print getConfigDict()
    print getAnnotationsConfigDict()
    print getLocalAttributesConfigDict()




#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
textParser.py
=============

A set of functions to parse text file data into lists, strings,
reals, integers etc.

"""

# Standard library imports
import re
import string

# Local imports
from nappy.utils.right_strip import *

# Global variables
pattnNoQuotes = re.compile("^[\"'].*\1$")


def readItemFromLine(line, rttype=str):
    """
    Reads an item of type ``rttype`` from ``line``.
    """
    line = rightStripCurlyBraces(line)
    rtitem = pattnNoQuotes.sub("", line.strip())

    if rttype is not str:
        rtitem = rttype(rtitem)   
    return rtitem

def readItemsFromLine(line, nitems=None, rttype=str):
    """
    Reads ``nitems`` items of type ``rttype`` from ``line``.
    """
    line = rightStripCurlyBraces(line)
    rtitems = re.split(r"\s+", line.strip())

    if nitems and len(rtitems) != nitems:
        raise "Incorrect number of items (%s) found in line: \n'%s'" % (nitems, line)
    if rttype is not str:
        rtitems = [rttype(x) for x in rtitems]
    return rtitems

def readItemsFromLines(lines, nitems, rttype=str):
    """
    Reads ``nitems`` items of type ``rttype`` from ``lines``
    """
    lines = [rightStripCurlyBraces(line) for line in lines]
    rtitems = []
    for line in lines:
        rtitems = rtitems + [readItemFromLine(line, rttype)]
    if rttype is not str:
        rtitems = [rttype(x) for x in rtitems]
    return rtitems

def readItemsFromUnknownLines(object, nitems, rttype=str):
    """
    Reads from an unknown number of lines until n items have been collected.
    The 'object' argument can be a filehandle (i.e. obj=open('name.ext', 'r'))
    or a string wrapped in a StringIO object (i.e. obj=StringIO.StringIO('abc')).
    The 'object' argument can also be a list, in which case the partially used/read object is
    also returned.
    """

    rtitems = []
    lines = []        

    if type(object) == type([2,3]): 

        while len(rtitems) < nitems:   
	    nextitem = object[0]
	    object = object[1:]
            items = rightStripCurlyBraces(nextitem).strip().split()
            lines.append(items)
            (rtitems, extras) = (rtitems + items[:nitems], items[nitems:])

    else:
        while len(rtitems) < nitems:
            items = rightStripCurlyBraces(object.readline()).strip().split()
            lines.append(items)
            (rtitems, extras) = (rtitems + items[:nitems], items[nitems:])

    if len(extras) > 0:
        raise Exception("Could not split " + `len(lines)` + " lines exactly into required number (" + `nitems` + ") of items: \n" + str(lines))

    if rttype is not str:
        rtitems = [rttype(x) for x in rtitems]

    if type(object) == type([1,2]):
        return (rtitems, object)
    else:
        return rtitems



#!/usr/bin/env python
#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
na2nc.py
========

Converts a NASA Ames file to a NetCDF file.

Usage
=====

   na2nc.py [-m <mode>] [-g <global_atts_list>]
            [-r <rename_vars_list>] [-t <time_units>] [-n] 
            -i <na_file> [-o <nc_file>] 

Where
-----

    <mode>			is the file mode, either "w" for write or "a" for append
    <global_atts_list>		is a comma-separated list of global attributes to add
    <rename_vars_list>		is a comma-separated list of <old_name>,<new_name> pairs to rename variables
    <time_units>		is a valid time units string such as "hours since 2003-04-30 10:00:00"
    -n				suppresses the time units warning if invalid
    <na_file>			is the input NASA Ames file path
    <nc_file>			is the output NetCDF file path (default is to replace ".na" from NASA Ames
           			 file with ".nc").
 
"""

# Imports from python standard library
import sys
import getopt

# Import from nappy package
import nappy
from nappy.utils.common_utils import makeListFromCommaSepString, makeDictFromCommaSepString


def exitNicely(msg=""):
    "Exits nicely!"
    print __doc__
    if msg != "": print "ERROR:", msg
    sys.exit()


def parseArgs(args):
    """
    Parses arguments and returns dictionary.
    """
    arg_dict = {}
    a = arg_dict

    # Set up defaults
    a["na_file"] = None
    a["mode"] = "w"
    a["variables"] = "all"
    a["aux_variables"]  = "all"
    a["global_attributes"] = {}
    a["rename_variables"] = {}
    a["time_units"] = None
    a["time_warning"] = True
    a["nc_file"] = None

    try:
        (arg_list, dummy) = getopt.getopt(args, "i:o:m:v:a:g:t:nr:")
    except getopt.GetoptError, e:
        exitNicely(str(e))
    
    for arg, value in arg_list:
        if arg == "-i":
            a["na_file"] = value
        elif arg == "-o":
            a["nc_file"] = value
        elif arg == "-m":
            a["mode"] = value
        elif arg == "-v":
            a["variables"] = value.split(",")
            exitNicely("OPTION NOT IMPLEMENTED: %s : Not yet fully functional. Sorry." % arg)
        elif arg == "-a":
            a["aux_variables"] = value.split(",")
            exitNicely("OPTION NOT IMPLEMENTED: %s : Not yet fully functional. Sorry." % arg)
        elif arg == "-g":
            a["global_attributes"] = makeListFromCommaSepString(value)
        elif arg == "-t":
            a["time_units"] = value
        elif arg == "-n":
            a["time_warning"] = False
        elif arg == "-r":
            a["rename_variables"] = makeDictFromCommaSepString(value)
        else:
            exitNicely("Argument '" + arg + "' not recognised!")

    if not a["na_file"]:
        exitNicely("Please provide argument '-i <na_file>'")

    if not a["nc_file"]:
        fn = a["na_file"]
        if fn[-3:] == ".na": fn = fn[:-3]
        nc_file = fn + ".nc"
        print "Auto-generating output file name:", nc_file
        a["nc_file"] = nc_file

    return a


def na2nc(args=None):
    """
    Controller for conversion of NASA Ames file to NetCDF file.
    """

    if args is None:
        args = sys.argv[1:]
    
    arg_dict = parseArgs(args)
    na_file = arg_dict["na_file"]
    del arg_dict["na_file"]
    nc_file = apply(nappy.convertNAToNC, [na_file], arg_dict)


if __name__ == "__main__":

    args=sys.argv[1:]
    na2nc(args)



#!/usr/bin/env python

"""
right_strip.py
==============

Holds the rightStripCurlyBraces() function used to right strip any curly braces 
annotations from lines in a text file.

E.g.:

In:  36   1001                              {NLHEAD  FFI}
Out: 36   1001

In:  Tout (C)            {Out-boarding air temperature}
Out: Tout (C) 

"""

import re

rstrip_regex = re.compile("^(.+)\{[^{^}]*\}\s*$")

def rightStripCurlyBraces(line):
    """
    Returns line but with curly braces right stripped off.
    """
    match = rstrip_regex.match(line)
    if not match: return line

    return match.groups()[0]


if __name__ == "__main__":

    test_lines = """36   1001                              {NLHEAD  FFI}
MDB {remove me}
34,56,23 {ddfsd {remove me now}
{h3lo} {world}    {IVOL NVOL}   	 
1 1 1 1 1 1 1 1 1  {lots of }  2 2 2 2 2 2      {NUMBERS}
trouble { { sdfsd } } } {} eat food {...} {...}
fairy god {mother} 						{hello
traffic 2k  2.71  5.32  -14.22 6.55   204.85  -0.41   67.8305 {{{don't worry}}}
will this {erase}}
 21                                    {NV}
 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
Tout (C)            {Out-boarding air temperature}
Uwind (m/s)         {Absolute value of wind velocity}
1                                      {NNCOML}""".split("\n")

    for line in test_lines:
        print "INPUT:  %s" % line
        print "OUTPUT: %s" % rightStripCurlyBraces(line)





#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
na_content_collector.py
=======================

Holds the class NAContentCollector that converts a set of CDMS variables and global attributes to a NASA Ames dictionary.

"""

# Imports from python standard library
import sys
import time
import re
import logging

# Import from nappy package
from nappy.na_error import na_error
import nappy.utils
import cdms_utils.axis_utils
import cdms_utils.var_utils
import nappy.utils.common_utils
import nappy.na_file.na_core

config_dict = nappy.utils.getConfigDict()
nc_to_na_map = config_dict["nc_to_na_map"]
header_partitions = config_dict["header_partitions"]
hp = header_partitions

version = nappy.utils.getVersion()

# Import external packages (if available)
if sys.platform.find("win") > -1:
    raise na_error.NAPlatformError("Windows does not support CDMS. CDMS is required to convert to CDMS objects and NetCDF.")
try:
    import cdms2 as cdms
    import numpy as N
except:
    try:
        import cdms
        import Numeric as N
    except:
        raise Exception("Could not import third-party software. Nappy requires the CDMS and Numeric packages to be installed to convert to CDMS and NetCDF.")

cdms.setAutoBounds("off") 
DEBUG = nappy.utils.getDebug() 

logging.basicConfig()
log = logging.getLogger(__name__)

class NAContentCollector(nappy.na_file.na_core.NACore):
    """
    Class to build a NASA Ames File object from a set of 
    CDMS variables and global attributes (optional).
    """
    
    def __init__(self, variables, global_attributes=[], requested_ffi=None):
        """
        Sets up instance variables and calls appropriate methods to
        generate sections of NASA Ames file object.

        Input arguments are:
          * variables - list/tuple of actual CDMS variables
          * global_attributes - list of user-defined global (key,value) attributes to include.

        Typical usage:
        >>> x = NAContentCollector(["temp", "precip"])
        >>> x.collectNAContent()
        >>> if x.found_na == True:
        ...     print x.na_dict, x.var_ids, x.unused_vars 
        """
        self.output_message = []
        self.na_dict = {}
        self.vars = variables

        # Note that self.var_ids will be a list containing:
        #    [ordered_vars,  auxiliary_vars,   rank_zero_vars]
        self.var_ids = None
        self.globals = dict(global_attributes)
        self.requested_ffi = requested_ffi

        self.rank_zero_vars = []
        self.rank_zero_var_ids = []

        # Create a flag to check if anything found
        self.found_na = False


    def collectNAContent(self):
        """
        Collect NASA Ames content. Save the contents to the following instance
        attributes:
         * self.na_dict
         * self.var_ids
         * self.unused_vars
        """
        log.debug("Call to collectNAContent():\n")
        for v in self.vars: 
            log.debug("\t%s, %s, %s" % (v.id, v.shape, v.getAxisIds()))
 
        (self.ordered_vars, aux_vars) = self._analyseVariables()
     
        if self.ordered_vars == []:
            log.warn("No NASA Ames content created.")
            self.unused_vars = []
        else:
            self.var_ids = [[var.id for var in self.ordered_vars],
                            [var.id for var in aux_vars], 
                            self.rank_zero_var_ids]
            self.na_dict["NLHEAD"] = -999
            self._defineNAVars(self.ordered_vars)
            self._defineNAAuxVars(aux_vars)
            self._defineNAGlobals()
            self._defineNAComments()
            self._defineGeneralHeader()
            self.found_na = True


    def _analyseVariables(self):
        """
        Method to examine the content of CDMS variables to return
        a tuple of two lists containing variables and auxiliary variables
        for the NASA Ames file object.
        Variables not compatible with the first file are put in self.unused_vars
        """
        self.unused_vars = []
        ffis_limited = False

        highest_rank = -1
        best_var = None
        count = 0

        # Need to get highest ranked variable (most dimensions) so that we can work out FFI
        for var in self.vars:
            msg = "Analysing: %s" % var.id
            self.output_message.append(msg)
            count = count + 1
            # get rank
            rank = var.rank()

            # Deal with singleton variables
            if rank == 0: 
                self.rank_zero_vars.append(var)
                self.rank_zero_var_ids.append(var.id)
                continue

            # Update highest if highest found or if equals highest with bigger size
            try:
                var.size = var.size() ; best_var.size = best_var.size()
            except:
                pass
            if rank > highest_rank or (rank == highest_rank and var.size > best_var.size):
                highest_rank = rank
                best_var = var
                best_var_index = count - 1

        # If all are zero ranked variables or no vars identified/found then we cannot write any to NASA Ames and return ([], [])
        if len(self.rank_zero_vars) == len(self.vars) or best_var is None: 
            return ([], [])

        # Now start to sort the variables into main and auxiliary 
        vars_for_na = [best_var]
        aux_vars_for_na = []
        shape = best_var.shape
        number_of_dims = len(shape)
        self.na_dict["NIV"] = number_of_dims


        # If 2D then do a quick test to see if 2310 is feasible (i.e. uniformly spaced 2nd axis)
        if number_of_dims == 2:
            ffis_limited = [2010, 2110]
            axis = best_var.getAxis(1)
            if cdms_utils.axis_utils.isUniformlySpaced(axis):
                ffis_limited.append(2310)

        # Get the axes for the main variable being used
        best_var_axes = best_var.getAxisList()
        
        # Get other variables into a list and analyse them
        rest_of_the_vars = self.vars[:best_var_index] + self.vars[(best_var_index + 1):]

        for var in rest_of_the_vars:

            if var.id in self.rank_zero_var_ids: continue

            # What to do with variables that have different number of dimensions or different shape
            if len(var.shape) != number_of_dims or var.shape != shape: 
                # Could it be an auxiliary variable?
                if len(var.shape) != 1: 
                    self.unused_vars.append(var)
                    continue

                first_axis = var.getAxis(0)
                # Check if axis is identical to first axis of main best variable, if so, can be auxiliary var
                if cdms_utils.axis_utils.areAxesIdentical(best_var_axes[0], first_axis) == False: 

                    # If not identical, then it might still qualify as an auxiliary every n time points - valid for 1020
                    if len(var.shape) == 1:
                        nvpm = cdms_utils.axis_utils.isAxisRegularlySpacedSubsetOf(first_axis, best_var_axes[0])
                        # NVPM is the number of implied values which is equal to (len(ax2)/len(ax1))
                        if nvpm:
                            ffis_limited = [1020]
                            self.na_dict["NVPM"] = nvpm
                        else: # if returned False, i.e. not regular subset axis
                            self.unused_vars.append(var)
                    else:
                        self.unused_vars.append(var)
                        continue

                else:
                    # This could be used as a standard auxiliary variable
                    if ffis_limited in ([1020],):
                        # Already fixed on 1020 and cannot collect incompatible FFI vars so do not use
                        self.unused_vars.append(var)
                    else:
                        aux_vars_for_na.append(var) 

            else:
                this_var_axes = var.getAxisList()

                # Loop through dimensions
                for i in range(number_of_dims):            
                    if cdms_utils.axis_utils.areAxesIdentical(best_var_axes[i], this_var_axes[i]) == False:
                        self.unused_vars.append(var)
                        break
                else:
                    # OK, I think the current variable is compatible to write with the best variable along with a NASA Ames file 
                    vars_for_na.append(var)

        # Send vars_for_na AND aux_vars_for_na to a method to check if they have previously been mapped 
        # from NASA Ames. In which case we'll write them back in the order they were initially read from the input file.
        (vars_for_na, aux_vars_for_na) = self._reorderVarsIfPreviouslyNA(vars_for_na, aux_vars_for_na)

        # Get the FFI
        self.na_dict["FFI"] = self._decideFileFormatIndex(number_of_dims, aux_vars_for_na, ffis_limited)
        return (vars_for_na, aux_vars_for_na)


    def _reorderVarsIfPreviouslyNA(self, vars_for_na, aux_vars_for_na):
        """
        Re-order if they previously came from NASA Ames files (i.e. including the 
        attribute 'nasa_ames_var_number'). Return re-ordered or unchanged pair of
        (vars_for_na, aux_vars_for_na).
        """
        # THIS SHOULD REALLY BE DONE IN A LOOP
        # First do the main variables
        ordered_vars = [None] * 1000 # Make a long list to put vars in 
        # Create a list of other variables to collect up any that are not labelled as nasa ames variables
        other_vars = []
        for var in vars_for_na:
            if hasattr(var, "nasa_ames_var_number"):
                ordered_vars[var.nasa_ames_var_number[0]] = var
            else:
                other_vars.append(var)

        # Remake vars_for_na now in new order and clean out any that are "None"
        vars_for_na = []
        for var in ordered_vars:
            if type(var) != type(None): 
                vars_for_na.append(var)

        vars_for_na = vars_for_na + other_vars

        # Now re-order the Auxiliary variables if they previously came from NASA 
        ordered_aux_vars = [None] * 1000
        other_aux_vars = []

        for var in aux_vars_for_na:
            if hasattr(var, "nasa_ames_aux_var_number"):
                ordered_aux_vars[var.nasa_ames_aux_var_number[0]] = var
            else:
                other_aux_vars.append(var)

        # Remake aux_vars_for_na now in order
        aux_vars_for_na = []
        for var in ordered_aux_vars:
            if type(var) != type(None): 
                aux_vars_for_na.append(var)

        aux_vars_for_na = aux_vars_for_na + other_aux_vars
        return (vars_for_na, aux_vars_for_na)


    def _decideFileFormatIndex(self, number_of_dims, aux_vars_for_na, ffis_limited=False):
        """
        Based on the number of dimensions and the NASA Ames dictionary return
        the File Format Index. 
        If there is a choice then make the most sensible selection.
        If the user has specified a 'requested_ffi' then try and deliver
        that. Raise an error if not possible.
        """
        # If ffis_limited is set then must use one of those
        if self.requested_ffi and ffis_limited:
            if self.requested_ffi not in ffis_limited:
                raise Exception("Cannot write this data to FFI '" + str(self.requested_ffi) + "', can only write to: " + str(ffis_limited) + ".")
            else:
                return self.requested_ffi

        # Base the sub-selection on number of dimensions
        if number_of_dims > 4:
            raise Exception("Cannot write variables defined against greater than 4 axes in NASA Ames format.")
        elif number_of_dims > 2: 
            ffi = 10 + (number_of_dims * 1000)
        elif number_of_dims == 2:
            if self.requested_ffi in (2010, 2110, 2310):
                ffi = self.requested_ffi 
            else:
                ffi = 2010
        else:
            if len(aux_vars_for_na) > 0 or (self.na_dict.has_key("NAUXV") and self.na_dict["NAUXV"] > 0):
                ffi = 1010
            else:
                ffi = 1001

        if self.requested_ffi and ffi != self.requested_ffi:
            raise Exception("Cannot write this data to FFI '" + str(self.requested_ffi) + "', can only write to: " + str(ffi) + ".")
        return ffi


    def _defineNAVars(self, vars):
        """
        Method to define NASA Ames file object variables and their
        associated metadata.
        """
        self.na_dict["NV"] = len(vars)
        self.na_dict["VNAME"] = []
        self.na_dict["VMISS"] = []
        self.na_dict["VSCAL"] = []
        self.na_dict["V"] = []

        for var in vars:
            name = cdms_utils.var_utils.getBestName(var)
            self.na_dict["VNAME"].append(name)
            miss = cdms_utils.var_utils.getMissingValue(var)

            if type(miss) not in (type(1.2), type(1), type(1L)):  
                miss = miss[0]

            self.na_dict["VMISS"].append(miss)
            self.na_dict["VSCAL"].append(1)

            # Populate the variable list with the array
            # Make sure missing values are converted to real values using the required missing value
            self.na_dict["V"].append(self._getFilledArrayAsList(var, miss))

            # Create independent variable info
            if not self.na_dict.has_key("X"):
                # Set up lists ready to populate with values
                self.na_dict["NXDEF"] = []
                self.na_dict["NX"] = []

                self.ax0 = var.getAxis(0)

                self.na_dict["X"] = [self.ax0[:].tolist()]
                self.na_dict["XNAME"] = [cdms_utils.var_utils.getBestName(self.ax0)]

                if len(self.ax0) == 1:
                    self.na_dict["DX"] = [0]
                else:
                    incr = self.ax0[1] - self.ax0[0]
                    # Set default increment as gap between first two
                    self.na_dict["DX"] = [incr]
                    # Now overwrite it as zero if non-uniform interval in axis
                    for i in range(1, len(self.ax0)):
                        if (self.ax0[i] - self.ax0[i - 1]) != incr:
                            self.na_dict["DX"] = [0]
                            break

                # If 1D only then "X" should only be a list and not list of lists
                if self.na_dict["FFI"] in (1001, 1010, 1020):
                    self.na_dict["X"] = self.na_dict["X"][0]

                # If FFI is 1020 need to reduce axis down to reduced values as most are implied
                if self.na_dict["FFI"] == 1020: 
                    vals = self.na_dict["X"]
                    self.na_dict["X"] = vals[0:len(vals):self.na_dict["NVPM"]] 

                # Now add the rest of the axes to the self.na_dict objects 
                for axis in var.getAxisList()[1:]:
                    self._appendAxisDefinition(axis)

                # If FFI is 2110 then need to modify the "NX" and "X" lists to cope with odd shape
                # Also need to add NX to auxiliary variables
                if self.na_dict["FFI"] == 2110:
                    new_x = []
                    new_nx = []
                    ax2_values = var.getAxis(1)[:].tolist()

                    for i in self.ax0[:]:
                        new_x.append([i, ax2_values])
                        new_nx.append(len(ax2_values))

                    # Re-assign to new lists
                    self.na_dict["NX"] = new_nx
                    self.na_dict["X"] = new_x                    

                    # Now auxiliary variable info here with independent var info
                    # First aux var is NX
                    self.na_dict["A"] = [self.na_dict["NX"][:]]
                    ind_var_name = self.na_dict["XNAME"][0]
                    self.na_dict["ANAME"] = ["Number of '%s' values recorded in subsequent data records" % ind_var_name]
                    self.na_dict["AMISS"] = [-9999.999]
                    self.na_dict["ASCAL"] = [1.0]

                # If FFI is 2310 then need to modify na_dict items for that
                elif self.na_dict["FFI"] == 2310:
                    new_x = []
                    new_nx = []
                    new_dx = []
                    ax2_values = var.getAxis(1)[:].tolist()
                    incr = ax2_values[1] - ax2_values[0]

                    for i in self.ax0[:]:
                        new_x.append([i, ax2_values])
                        new_nx.append(len(ax2_values))
                        new_dx.append(incr)

                    # Re-assign to new lists
                    self.na_dict["NX"] = new_nx
                    self.na_dict["X"] = new_x
                    self.na_dict["DX"] = new_dx

                    # Now auxiliary variable info here with independent var info
                    # First three aux vars are NX, X0 and DX
                    self.na_dict["A"] = []
                    self.na_dict["A"].append(self.na_dict["NX"][:])
                    self.na_dict["A"].append([i[1][0] for i in self.na_dict["X"]])
                    self.na_dict["A"].append(self.na_dict["DX"][:])
                    ind_var_name = self.na_dict["XNAME"][0]
                    self.na_dict["ANAME"] = ["Number of '%s' values recorded in subsequent data records" % ind_var_name,
                                             "'%s' value for first data point" % ind_var_name,
                                             "'%s' increment" % ind_var_name]
                    self.na_dict["AMISS"] = [-9999.999, -9999.999, -9999.999]
                    self.na_dict["ASCAL"] = [1.0, 1.0, 1.0]
 

    def _defineNAAuxVars(self, aux_vars):
        """
        Method to define NASA Ames file object auxiliary variables and their
        associated metadata. Note that "A" may already have content if
        independent variable items (relating to "X") are defined as aux vars.
        """
        # Initialise aux var itesms as empty lists unless already defined when
        # setting up independent variables
        for item in ("ANAME", "AMISS", "ASCAL", "A"):
            if not self.na_dict.has_key(item):
                self.na_dict[item] = [] 

        for var in aux_vars:
            name = cdms_utils.var_utils.getBestName(var)
            self.na_dict["ANAME"].append(name)
            miss = cdms_utils.var_utils.getMissingValue(var)
            if type(miss) not in (type(1.1), type(1), type(1L)):  miss = miss[0]
            self.na_dict["AMISS"].append(miss)
            self.na_dict["ASCAL"].append(1)
            # Populate the variable list with the array
            self.na_dict["A"].append(self._getFilledArrayAsList(var, miss))

        self.na_dict["NAUXV"] = len(self.na_dict["A"])

    def _appendAxisDefinition(self, axis):
        """
        Method to create the appropriate NASA Ames file object 
        items associated with an axis (independent variable in 
        NASA Ames). It appends to the various self.na_dict containers.
        """
        length = len(axis)

        self.na_dict["NX"].append(length)
        self.na_dict["XNAME"].append(cdms_utils.var_utils.getBestName(axis))
        # If only one item in axis values
        if length < 2:
            self.na_dict["DX"].append(0)
            self.na_dict["NXDEF"].append(length)
            self.na_dict["X"].append(axis[:].tolist())        
            return
   
        incr = axis[1] - axis[0]
        for i in range(1, length):
            if (axis[i] - axis[i - 1]) != incr:
                self.na_dict["DX"].append(0)
                self.na_dict["NXDEF"].append(length)
                self.na_dict["X"].append(axis[:].tolist())
                break
        else: # If did not break out of the loop
            max_length = length
            if length > 3: 
                max_length = 3
            self.na_dict["DX"].append(incr)
            self.na_dict["NXDEF"].append(max_length)
            self.na_dict["X"].append(axis[:max_length])

    def _defineNAGlobals(self):
        """
        Maps CDMS (NetCDF) global attributes into NASA Ames Header fields.
        """
        # Check if we should add to it with locally set rules
        local_attributes = nappy.utils.getLocalAttributesConfigDict()
        local_nc_atts = local_attributes["nc_attributes"]
         
        for att, value in local_nc_atts.items():
            if not nc_to_na_map.has_key(att):
                nc_to_na_map[key] = value

        self.extra_comments = [[],[],[]]  # Normal comments, special comments, other comments
        convention_or_reference_comments = []

        for key in self.globals.keys():
            if key != "first_valid_date_of_data" and type(self.globals[key]) \
                                       not in (type("s"), type(1.1), type(1)):
                continue

            # Loop through keys of header/comment items to map
            if key in nc_to_na_map.keys():
                if key == "history":
                    time_string = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
                    history = "History:  %s - Converted to NASA Ames format using nappy-%s.\n  %s" % \
                                                 (time_string, version, self.globals[key])
                    history = history.split("\n") 
                    self.history = []
                    for h in history:
                        if h[:8] != "History:" and h[:1] != "  ": 
                            h = "  " + h
                        self.history.append(h) 
                    
                elif key == "institution":
                    # If fields came from NA then extract appropriate fields.
                    match = re.match(r"(.*)\s+\(ONAME from NASA Ames file\);\s+(.*)\s+\(ORG from NASA Ames file\)\.", 
                             self.globals[key])
                    if match:
                        self.na_dict["ONAME"] = match.groups()[0]
                        self.na_dict["ORG"] = match.groups()[1]
                    else:
                        self.na_dict["ONAME"] = self.globals[key]
                        self.na_dict["ORG"] = self.globals[key]
                    # NOTE: should probably do the following search and replace on all string lines
                    self.na_dict["ONAME"] = self.na_dict["ONAME"].replace("\n", "  ")
                    self.na_dict["ORG"] = self.na_dict["ORG"].replace("\n", "  ")

                elif key == "comment":
                    # Need to work out if they are actually comments from NASA Ames in the first place
                    comment_lines = self.globals[key].split("\n")
                    normal_comments = []
                    normal_comm_flag = None
                    special_comments = []
                    special_comm_flag = None

                    for line in comment_lines:
                        if line.find(hp["sc_start"]) > -1:
                            special_comm_flag = 1
                        elif line.find(hp["sc_end"]) > -1:
                            special_comm_flag = None
                        elif line.find(hp["nc_start"]) > -1:
                            normal_comm_flag = 1
                        elif line.find(hp["nc_end"]) > -1:
                            normal_comm_flag = None
                        elif special_comm_flag == 1:
                            special_comments.append(line)
                        elif normal_comm_flag == 1:
                            normal_comments.append(line)
                        elif line.find(hp["data_next"]) > -1:
                            pass
                        else:
                            normal_comments.append(line)    

                    self.extra_comments = [special_comments, normal_comments, []]    

                elif key == "first_valid_date_of_data":
                    self.na_dict["DATE"] = self.globals[key]

                elif key in ("Conventions", "references"):
                    #convention_or_reference_comments.append("%s:   %s" % (key, self.globals[key]))
                    self.extra_comments[2].append("%s:   %s" % (key, self.globals[key]))
                else:
                    self.na_dict[nc_to_na_map[key]] = self.globals[key]
            else:
                self.extra_comments[2].append("%s:   %s" % (key, self.globals[key]))
        return


    def _defineNAComments(self, normal_comments=[], special_comments=[]):
        """
        Defines the Special and Normal comments sections in the NASA Ames file 
        object - including information gathered from the defineNAGlobals method.

        Starts with values provided for normal_comments and special_comments.
        """
        if hasattr(self, "ncom"):  normal_comments = self.ncom + normal_comments

        NCOM = []
        for ncom in normal_comments:
            NCOM.append(ncom)

        if len(NCOM) > 0:   NCOM.append("")

        # Use third item in self.extra_comments and adds to NCOM
        if len(self.extra_comments[2]) > 0:
            for excom in self.extra_comments[2]:
                NCOM.append(excom)

        if len(self.extra_comments[1]) > 0:  
            NCOM.append(hp["addl_globals"])
            for excom in self.extra_comments[1]:
                NCOM.append(excom)

        if hasattr(self, "history"):
            for h in self.history:
                NCOM.append(h)
       
        # When NCOM has been defined then surround it in some extras 
        if len(NCOM) > 0:
            NCOM.insert(0, hp["nc_start"]) 
            NCOM.append("")
            NCOM.append(hp["nc_end"])
            NCOM.append(hp["data_next"])

        spec_comm_flag = None
        # Start with special_comments added in
        SCOM = []

        # Uses first item in self.extra_comments to start SCOM
        special_comments = special_comments + self.extra_comments[0]

        if len(special_comments) > 0: 
            SCOM = [hp["sc_start"]]
            spec_comm_flag = 1

        for scom in special_comments:
            SCOM.append(scom)

        used_var_atts = ("id",  "missing_value", "fill_value", 
                   "nasa_ames_var_number", "nasa_ames_aux_var_number")
        var_comm_flag = None

        # Create a string for the Special comments to hold rank-zero vars
        rank_zero_vars_string = []

        for var in self.rank_zero_vars:
            rank_zero_vars_string.append("  Variable %s: %s" % (var.id, cdms_utils.var_utils.getBestName(var)))

            for att in var.attributes.keys():
                value = var.attributes[att]

                if type(value) in (type("s"), type(1.0), type(1)):

                    rank_zero_vars_string.append("    %s = %s" % (att, var.attributes[att]))

        if len(rank_zero_vars_string) > 0:
            rank_zero_vars_string.insert(0, hp["sing_start"])
            rank_zero_vars_string.append(hp["sing_end"])

        for var in self.ordered_vars:
            varflag = "unused"
            var_name_written = False

            name = cdms_utils.var_utils.getBestName(var)

            for scom,value in var.attributes.items():
                if type(value) in (type([]), type(N.array([0]))) and len(value) == 1:
                    value = value[0]

                if type(value) in (type("s"), type(1.1), type(1)) and scom not in used_var_atts:
                    if varflag == "unused":
                        if var_comm_flag == None:
                            var_comm_flag = 1

                    if spec_comm_flag == None:
                        SCOM = [hp["sc_start"]] + rank_zero_vars_string
                        SCOM.append(hp["addl_vatts"])
                        SCOM.append(hp["ncatts_start"])
                        varflag = "using" 
                        spec_comm_flag = 1

                    if var_name_written == False:
                        SCOM.append("  Variable %s: %s" % (var.id, name))
                        var_name_written = True

                    SCOM.append("    %s = %s" % (scom, value))

        if var_comm_flag == 1:  
            SCOM.append(hp["ncatts_end"])
        if spec_comm_flag == 1:
            SCOM.append(hp["sc_end"])

        # Strip out empty lines (or returns)
        NCOM_cleaned = []
        SCOM_cleaned = []

        for c in NCOM:
            if c.strip() not in ("", " ", "  "):
                # Replace new lines within one attribute with a newline and tab so easier to read
                lines = c.split("\n")
                for line in lines:
                    if line != lines[0]: 
                        line = "  " + line

                    NCOM_cleaned.append(line)

        for c in SCOM:
            if c.strip() not in ("", " ", "  "):
                # Replace new lines within one attribute with a newline and tab so easier to read
                lines = c.split("\n")
                for line in lines:
                    if line != lines[0]: 
                        line = "  " + line

                    SCOM_cleaned.append(line)

        self.na_dict["NCOM"] = NCOM_cleaned
        self.na_dict["NNCOML"] = len(self.na_dict["NCOM"])
        self.na_dict["SCOM"] = SCOM_cleaned
        self.na_dict["NSCOML"] = len(self.na_dict["SCOM"])
        return


    def _defineGeneralHeader(self, header_items={}):
        """
        Defines known header items and overwrites any with header_items 
        key/value pairs.
        """
        warning_message = "Nappy Warning: Could not get the first date in the file. You will need to manually edit the output file."

        # Check if DATE field previously known in NASA Ames file
        time_now = [int(i) for i in time.strftime("%Y %m %d", time.localtime(time.time())).split()]

        if not self.na_dict.has_key("RDATE"):
            self.na_dict["RDATE"] = time_now

        if self.ax0.isTime():
            # Get first date in list
            try:
                (unit, start_date) = re.match("(\w+)\s+?since\s+?(\d+-\d+-\d+)", self.ax0.units).groups()            
                comptime = cdtime.s2c(start_date)
                first_day = comptime.add(self.na_dict["X"][0], getattr(cdtime, unit.capitalize()))
                self.na_dict["DATE"] = [int(i) for i in str(first_day).split(" ")[0].replace("-", " ").split()]
            except:
                msg = warning_message
                log.info(msg)
                self.output_message.append(msg)
                self.na_dict["DATE"] = [999] * 3 
        else: 
            if not self.na_dict.has_key("DATE"):
                msg = warning_message
                log.info(msg)
                self.output_message.append(msg)
                self.na_dict["DATE"] = [999] * 3 
            else:
                pass # i.e. use existing DATE

        self.na_dict["IVOL"] = 1
        self.na_dict["NVOL"] = 1
        for key in header_items.keys():
             self.na_dict[key] = header_items[key]


    def _getFilledArrayAsList(self, arr, missing_value):
        """
        Takes an array ``arr`` (either a numpy array or a Masked Array).
        If the array is a masked array then replace masked values with 
        ``missing_value`` and convert to a numpy array.
        Finally convert to a list and return that.
        """
        if N.ma.isMaskedArray(arr):
            arr = arr.filled(missing_value)

        return arr[:].tolist()




"""
ffi_translator.py
=================

Provides a set of functions to do simple mappings between FFIs by altering the required 
parts of the NASA Ames dictionary (na_dict) sent to them.

A good example is the translation of 2010 --> 2110 that requires some re-jigging of the 
ancillary variables (to include NX in the auxiliary variable rows and X values in the 
main data blocks before the actual data).

"""

# Import standard library modules
import copy

# Import local modules
import nappy.utils

default_missing_value = nappy.utils.getConfigDict()["main"]["default_missing_value"]


def translate2010To2110(na_dict):
    """
    Takes a valid na_dict for 2010 and returns a valid na_dict
    object for 2110.
    NOTE: this is a constrained version of 2110 that does not include
    the "ragged array" structures that are allowed.
    """
    n = copy.deepcopy(na_dict)

    # First need to adjust auxiliary variable definitions and array
    nx1 = n.NX[0]
    n.NAUXV = n.NAUXV + 1
    axis_1_name = n.XNAMES[0]
    n.ANAMES = ["Number of %s values recorded in subsequent data records" % axis_1_name] + n.ANAMES
    n.ASCAL.insert(0, 1.0)
    n.AMISS.insert(0, default_missing_value)
   
    return n   # did it work?



 

    


#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
na_to_nc.py
===========

Contains the NAToNC class for converting a NASA Ames file to a NetCDF file.

"""

# Imports from python standard library
import logging

# Imports from external packages
try:
    import cdms2 as cdms
except:
    try:
        import cdms
    except:
        raise Exception("Could not import third-party software. Nappy requires the CDMS and Numeric packages to be installed to convert to CDMS and NetCDF.")


# Import from nappy package
import nappy.nc_interface.na_to_cdms
from nappy.na_error import na_error

logging.basicConfig()
log = logging.getLogger(__name__)

class NAToNC(nappy.nc_interface.na_to_cdms.NADictToCdmsObjects):
    """
    Converts a NASA Ames file to a NetCDF file.
    """
    
    def __init__(self, na_file, variables=None, aux_variables=None,
                 global_attributes=[("Conventions","CF-1.0")],
                 time_units=None, time_warning=True, 
                 rename_variables={}):
        """
        Sets up instance variables. Note that the argument 'na_file' has a relaxes definition
        and can be either a NASA Ames file object or the name of a NASA AMES file.
        Typical usage is:
        >>>    import nappy.nc_interface.na_to_nc as na_to_nc
        >>>    c = na_to_nc.NAToNC("old_file.na") 
        >>>    c.convert()
        >>>    c.writeNCFile("new_file.nc")         
        """
        # First open na_file if it is a file rather than an na_file object
        na_file_obj = na_file
        if type(na_file_obj) == type("string"):
            na_file_obj = nappy.openNAFile(na_file_obj)

        nappy.nc_interface.na_to_cdms.NADictToCdmsObjects.__init__(self, na_file_obj, variables=variables, 
                 aux_variables=aux_variables,
                 global_attributes=global_attributes,
                 time_units=time_units, time_warning=time_warning, 
                 rename_variables=rename_variables)


    def writeNCFile(self, file_name, mode="w"):
        """
        Writes the NASA Ames content that has been converted into CDMS objects to a
        NetCDF file of name 'file_name'. Note that mode can be set to append so you 
        can add the data to an existing file.
        """
        if not self.converted:
            self.convert()

        # Create CDMS output file object
        fout = cdms.open(file_name, mode=mode)

        # Write main variables
        for var in self.cdms_variables:
            fout.write(var)

        # Write aux variables
        for avar in self.cdms_aux_variables:
            fout.write(avar)

        # Write global attributes
        for (att, value) in self.global_attributes:
            setattr(fout, att, value)
        
        fout.close()
        log.info("NetCDF file '%s' written successfully." % file_name)
        return True



#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naCore.py
=========

Holds the NACore class that holds the None type version
of all the possible NASA Ames variables described in the Gaines
and Hipskind (1998)document. It also holds a number of useful methods
for accessing metadata within the file.

"""

# Imports from standard python library
import copy
import re

class NACore:
    """
    Abstract class to hold the empty NASA Ames contents and 
    a number of methods to access information in files. This
    class is sub-classed by all NAFile classes.
    """
    
    var_and_units_pattern = re.compile("^\s*(.*)\((.+?)\)(.*)\s*$")
    na_dictionary_keys = ("A", "AMISS", "ANAME", "ASCAL", "DATE", "DX",
                     "FFI", "IVOL", "LENA", "LENX", "MNAME", "NAUXC",
                     "NAUXV", "NCOM", "NIV", "NLHEAD", "NNCOML",
                     "NSCOML", "NV", "NVOL", "NVPM", "NX", "NXDEF",
                     "ONAME", "ORG", "RDATE", "SCOM", "SNAME", "V",
                     "VMISS", "VNAME", "VSCAL", "X", "XNAME")

    def __init__(self):
        """
        Creates an instance variable of every type used in all
        NA formats. All are set to None until/unless defined.
        """
        # Set up attributes for all possible NASA Ames dictionary items
        for key in NACore.na_dictionary_keys:
            setattr(self, key, None)

    def getNADict(self):
        """
        Returns a dictionary of the contents of a NASA Ames file.
        """
        dct = {}
        for key in NACore.na_dictionary_keys:
            dct[key] = getattr(self, key)

        self.na_dict={}
        for i in dct.keys():
            if dct[i] != None:
                self.na_dict[i] = dct[i]
        return self.na_dict

    def setNADict(self, na_dict):
        """
        Dynamic setting of the na_dict dictionary.
        """
        self.na_dict = copy.deepcopy(na_dict)

    def __getitem__(self, item):
        """
        Dictionary item access to NASA Ames contents, called NAFileObj['NIV']
        will return NAFileObj.NIV.
	
        Note: In future this might return whatever user wants and to translate
        NASA Ames variables such as 'NIV' to explanatory strings
        such as 'number_of_independent_variables'. Need a map for
        this defined at top of the nasaAmesData.py module
        """
        if hasattr(self, item):
            return getattr(self, item)
        else:
            return "Item '%s' not found." % item

    def _attemptVarAndUnitsMatch(self, item):
        """
        If it can match variable name and units from the name it does and returns
        (var_name, units). Otherwise returns (item, None).
        """
        match = NACore.var_and_units_pattern.match(item)

        if match:
            (v1, units, v2) = match.groups()
            var_name = v1 + " " + v2
        else:
            (var_name, units) = (item, None)   
  
        return (var_name.strip(), units)

    def getVariable(self, var_number): 
        """
        Returns variable metadata corresponding to the var_number argument in the 
        list of varibles. Tuple of (variable_name, units, missing_value, scale_factor)
        is returned.
        """
        (variable, units) = self._attemptVarAndUnitsMatch(self.VNAME[var_number])
        miss = self.getMissingValue(var_number)
        scale = self.getScaleFactor(var_number)
        return (variable, units, miss, scale)

    def getIndependentVariable(self, ivar_number):
        """
        Returns an independent variable name and units in a tuple corresponding to
        the ivar_number index in the list.
        """
        (variable, units) = self._attemptVarAndUnitsMatch(self.XNAME[ivar_number])
        return (variable, units)

    def getAuxVariable(self, avar_number):        
        """
        Returns an auxiliary variable name and units in a tuple corresponding to
        the ivar_number index in the list.
        """
        (variable, units) = self._attemptVarAndUnitsMatch(self.ANAME[avar_number])
        miss = self.getAuxMissingValue(avar_number)
        scale = self.getAuxScaleFactor(avar_number)
        return (variable, units, miss, scale)    

    def getVariables(self):
        """
        Returns metadata for all main (non-auxiliary or independent) variables.
        """
        vars=[]
        for i in range(self.NV):
            vars.append(self.getVariable(i))
        return vars

    def getIndependentVariables(self):
        """
        Returns metadata for all independent variables.
        """
        ivars=[]
        for i in range(self.NIV):
            ivars.append(self.getIndependentVariable(i))
        return ivars

    def getAuxVariables(self):
        """
        Returns metadata for all auxiliary variables.
        """
        avars = []
        if not hasattr(self, "NAUXV"):
            for i in range(self.NAUXV):
                avars.append(self.getAuxVariable(i))
        else:
            avars = []
        return avars

    def getMissingValue(self, var_number):
        """
        Returns a missing value for a given variable.
        """
        return self.VMISS[var_number]


    def getScaleFactor(self, var_number):
        """
        Returns a scale factor for a given variable.
        """
        return self.VSCAL[var_number]
	
    def getAuxMissingValue(self, avar_number):
        """
        Returns the missing value of an auxiliary variable.
        """
        return self.AMISS[avar_number]

    def getAuxScaleFactor(self, avar_number):
        """
        Returns the scale factor of an auxiliary variable.
        """ 
        return self.ASCAL[avar_number]

    def getNumHeaderLines(self):
        """
        Returns the number of header lines.
        """
        return self.NLHEAD

    def getFFI(self):
        """
        Returns the File Format Index for the file.
        """
        return self.FFI

    def getOriginator(self):
        """
        Returns the Originator (ONAME) string.
        """
        return self.ONAME

    def getOrganisation(self):
        """
        Returns the Organisation (ORG) string.
        """
        return self.ORG

    def getOrg(self):
        """
        Returns the Organisation (ORG) string.
        """
        return self.getOrganisation()

    def getSource(self):
        """
        Returns the Source (SOURCE) string.
        """
        return self.SNAME

    def getMission(self):
        """
        Returns the mission (MNAME) string.	
        """
        return self.MNAME

    def getVolumes(self):
        """
        Returns the volume numbers (IVOL and NVOL).
        """
        return (self.IVOL, self.NVOL)

    def getFileDates(self):
        """
        Returns the first valid date in the data (DATE) and the 
        Revision Date (RDATE).
        """
        return (self.DATE, self.RDATE)

    def getNormalComments(self):
        """
        Returns the Normal Comment (NCOM) lines.
        """
        return self.NCOM

    def getSpecialComments(self):
        """
        Returns the Special Comments (SCOM) lines.
        """
        return self.SCOM




#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
na_to_cdms.py
=============

Container module for class NADictToCdmsObjects that is sub-classed
by NAToNC classes.

"""

# Imports from python standard library
import sys
import re
import time
import logging

# Import from nappy package
from nappy.na_error import na_error
import nappy.utils
import nappy.utils.common_utils

config_dict = nappy.utils.getConfigDict()
na_to_nc_map = config_dict["na_to_nc_map"]
header_partitions = config_dict["header_partitions"]
hp = header_partitions

# Import external packages (if available)
if sys.platform.find("win") > -1:
    raise na_error.NAPlatformError("Windows does not support CDMS. CDMS is required to convert to CDMS objects and NetCDF.")

try:
    import cdms2 as cdms
    import numpy as N
except:
    try:
        import cdms
        import Numeric as N
    except:
        raise Exception("Could not import third-party software. Nappy requires the CDMS and Numeric packages to be installed to convert to CDMS and NetCDF.")


# Define global variables
safe_nc_id = re.compile("[\/\s\[\(\)\]\=\+\-\?\#\~\@\&\$\%\!\*\{\}\^]+")
time_units_pattn = re.compile("\w+\s+since\s+\d{4}-\d{1,2}-\d{1,2}\s+\d+:\d+:\d+")
max_id_length = 40
special_comment_known_strings = (hp["sc_start"], hp["sc_end"], hp["addl_vatts"],
                                  hp["addl_globals"], "\n")

normal_comment_known_strings = (hp["nc_start"], hp["nc_end"], hp["data_next"], 
			          hp["addl_vatts"], hp["addl_globals"], "\n")

time_units_warning_message = """\nWARNING: Could not recognise time units. For true NetCDF compability
please insert the correct time unit string below in the format:
    
    <units> since <YYYY>-<MM>-<DD> <hh>-<mm>-<ss>
    
Where:
    <units> is a known time interval such as years, months, days, etc.
    <YYYY> is the year, <MM> is the month, <DD> is the day,
    <hh> is the hour, <mm> is minutes, <ss> is seconds.
"""

DEBUG = nappy.utils.getDebug() 

logging.basicConfig()
log = logging.getLogger(__name__)

class NADictToCdmsObjects:
    """
    Converts a NA File instance to a tuple of CDMS objects.
    """
    
    def __init__(self, na_file_obj, variables="all", aux_variables="all",
                 global_attributes=[("Conventions", "CF-1.0")],
                 time_units=None, time_warning=True, 
                 rename_variables={}):
        """
        Sets up instance variables.
        """
        self.na_file_obj = na_file_obj	   
        self.variables = variables
        self.aux_variables = aux_variables
        self.global_attributes = global_attributes
        self.time_units = time_units
        self.time_warning = time_warning
        self.rename_variables = rename_variables

        # Check if we have capability to convert this FFI
        if self.na_file_obj.FFI in (2110, 2160, 2310): 
	        raise Exception("Cannot convert NASA Ames File Format Index (FFI) " + `self.na_file_obj.FFI` + " to CDMS objects. No mapping implemented yet.")

        self.output_message = []  # for output displaying message
        self.converted = False

    def convert(self):
        """
        Reads the NASA Ames file object and converts to CDMS objects.
        Returns (variable_list, aux_variable_list, global_attributes_list).
        All these can be readily written to a CDMS File object.
        """
        if self.converted == True:
            log.info("Already converted to CDMS objects so not re-doing.")
            return (self.cdms_variables, self.cdms_aux_variables, self.global_attributes)

        self.na_file_obj.readData()

        # Convert global attribute
        self._mapNACommentsToGlobalAttributes()

        # Convert axes
        if not hasattr(self, 'cdms_axes'):  self._convertCdmsAxes()

        # Convert main variables
        if not hasattr(self, 'cdms_variables'):  self._convertCdmsVariables()

        # Then do auxiliary variables
        if hasattr(self.na_file_obj, "NAUXV") and (type(self.na_file_obj.NAUXV) == type(1) and self.na_file_obj.NAUXV > 0):   # Are there any auxiliary variables?
            if not hasattr(self, 'cdms_aux_variables'):  
                self._convertCdmsAuxVariables()
        else:
            self.cdms_aux_variables = []
            
        self.converted = True
        return (self.cdms_variables, self.cdms_aux_variables, self.global_attributes)


    def _mapNACommentsToGlobalAttributes(self):
        """
        Maps the NASA Ames comments section to global attributes and append them to the 
        self.global_attributes list.
        """
        glob_atts = dict(self.global_attributes)
        
        for key in na_to_nc_map.keys():

            if type(key) == type((1,2)):

                if key == ("SCOM", "NCOM"):
                    # Map special and normal comments into the global comments section
                    comment_line = ""

                    # Add Special comments first
                    if self.na_file_obj.NSCOML > 0:
                        comment_line += (hp["sc_start"] + "\n")

                        for i in self.na_file_obj.SCOM: 
			    
                            if i.strip() not in special_comment_known_strings:
                                comment_line += ("\n" + i)
                            
                        comment_line += ("\n" + hp["sc_end"] + "\n")

                    # Now add normal comments
                    if self.na_file_obj.NNCOML > 0:
                        comment_line += (hp["nc_start"] + "\n")

                        for i in self.na_file_obj.NCOM: 
                            if i.strip() not in normal_comment_known_strings:
                                comment_line += ("\n" + i)

                        comment_line += ("\n" + hp["nc_end"])

                    # Tidy comment line then write to global atts dict
                    comment_line = comment_line.replace("\n\n", "\n")

                    glob_atts["comment"] = comment_line

                elif key == ("ONAME", "ORG"):
                    # Map the two organisation NA files to the institution field in CDMS (NetCDF)
                    institution = "%s (ONAME from NASA Ames file); %s (ORG from NASA Ames file)." % \
                                         (self.na_file_obj.ONAME, self.na_file_obj.ORG)
                    glob_atts["institution"] = institution

                else:
                    # Any other strange tuple just gets merged into a string
                    item = (getattr(self.na_file_obj, key[0])) + "\n" + (getattr(self.na_file_obj, key[1]))
                    glob_atts[na_to_nc_map[key]] = item

            elif key == "RDATE":
                # RDATE = Revision date - update this and put in history global attribute
                date_parts = getattr(self.na_file_obj, "RDATE")
                date_string = "%.4d-%.2d-%.2d" % tuple(date_parts)
                hist = date_string + " - NASA Ames File created/revised.\n"
                time_string = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
                version = nappy.utils.getVersion()
                hist = "%s\n%s - Converted to CDMS (NetCDF) format using nappy-%s." % (hist, time_string, version)
                # self.cdms_file.history = hist
                log.debug("No history mapping from na so added it here from global attributes.")
                glob_atts["history"] = hist           
            else:
                # Anything else just needs to be stored as a global attribute
                glob_atts[na_to_nc_map[key]] = getattr(self.na_file_obj, key)

        # Now remake global atts list
        new_atts = []

        for key, value in self.global_attributes:
            new_atts.append( (key, glob_atts[key]) )
        used_keys = [i[0] for i in new_atts]

        for key in glob_atts.keys():
            if key not in used_keys:
                new_atts.append( (key, glob_atts[key]) )

        self.global_attributes = new_atts[:]

    def _convertCdmsVariables(self):
        """
        Creates cdms variable list for writing out.
        """
        self.cdms_variables = []

        if self.variables in (None, "all"):    
            for var_number in range(self.na_file_obj.NV):
                self.cdms_variables.append(self._convertNAToCdmsVariable(var_number))
        else:
            if type(self.variables[0]) == type(1) or re.match("\d+", str(self.variables[0])): # They are integers = indices
                for var_number in self.variables:
                    vn = int(var_number)
                    self.cdms_variables.append(self._convertNAToCdmsVariable(vn))   
            elif type(self.variables[0]) == type("string"):  # Vars are strings
                for var_name in self.variables:
                    if var_name in self.na_file_obj.VNAME:
                        var_number = self.na_file_obj.VNAME.index(var_name)
                        self.cdms_variables.append(self._convertNAToCdmsVariable(var_number))
                    else:
                        raise Exception("Variable name not known: " + var_name)


    def _convertNAToCdmsVariable(self, var_number, attributes={}):
        """
        Creates a single cdms variable from the variable number provided in the list.
        """
        (var_name, units, miss, scal) = self.na_file_obj.getVariable(var_number)
        msg = "\nAdding variable: %s" % self.na_file_obj.VNAME[var_number]
        log.debug(msg)
        self.output_message.append(msg)
        array = N.array(self.na_file_obj.V[var_number])
        array = array * scal
        # Set up axes
        if not hasattr(self, 'cdms_axes'):
            self._convertCdmsAxes()

        # Set up variable
        var=cdms.createVariable(array, axes=self.cdms_axes, fill_value=miss, attributes=attributes)

        # Sort units etc
        if units:   var.units=units
	
        # Add the best variable name
        if len(var_name) < max_id_length:
            var.id=safe_nc_id.sub("_", var_name).lower()
        else:
            var.id="naVariable_%s" % (var_number)
        
	     # Check if mapping provided for renaming this variable
        if var_name in self.rename_variables.keys():
            var_name = self.rename_variables[var_name]
	    
        var.long_name = var.name = var.title = var_name

        # Add a NASA Ames variable number (for mapping correctly back to NASA Ames)
        var.nasa_ames_var_number = var_number
        return var

    def _convertCdmsAuxVariables(self):
        """
        Creates a cdms variable from an auxiliary variable
        """
        self.cdms_aux_variables = []

        if self.aux_variables in (None, "all"):
            for avar_number in range(self.na_file_obj.NAUXV):
                self.cdms_aux_variables.append(self._convertNAAuxToCdmsVariable(avar_number))
        else:
            if type(self.aux_variables[0]) == type(1): # They are integers = indices
                for avar_number in self.aux_variables:
                    self.cdms_aux_variables.append(self._convertNAAuxToCdmsVariable(avar_number))   

            elif type(self.aux_variables[0]) == type("string"): # They are strings
                for avar_name in self.aux_variables:
                    if avar_name in self.na_file_obj.ANAME:
                        avar_number = self.na_file_obj.ANAME.index(avar_name)
                        self.cdms_aux_variables.append(self._convertNAAuxToCdmsVariable(avar_number)) 
            else:
                raise Exception("Auxiliary variable name not known: " + avar_name)	    

    def _convertNAAuxToCdmsVariable(self, avar_number, attributes={}):
        """
        Converts an auxiliary variable to a cdms variable.
        """
        (var_name, units, miss, scal) = self.na_file_obj.getAuxVariable(avar_number)
        array = N.array(self.na_file_obj.A[avar_number])
        array = array * scal

        msg="\nAdding auxiliary variable: %s" % self.na_file_obj.ANAME[avar_number]
        log.debug(msg)
        self.output_message.append(msg)

        # Set up axes
        if not hasattr(self, 'cdms_axes'):
            self._convertCdmsAxes()

        # Set up variable
        var = cdms.createVariable(array, axes=[self.cdms_axes[0]], fill_value=miss, 
                                  attributes=attributes)

        # Sort units etc
        if units:   var.units = units
        if len(var_name) < max_id_length:
            var.id = safe_nc_id.sub("_", var_name).lower()
        else:
            var.id = "naAuxVariable_%s" % (avar_number)

	    # Check if mapping provided for renaming this variable
        if var_name in self.rename_variables.keys():
            var_name = self.rename_variables[var_name]

        var.long_name = var.name = var.title = var_name

        # Add a NASA Ames auxiliary variable number (for mapping correctly back to NASA Ames)
        var.nasa_ames_aux_var_number = avar_number
        return var        

    def _convertCdmsAxes(self): 
        """
        Creates cdms axes from information provided in the NASA Ames dictionary.
        """
        if not hasattr(self, 'cdms_axes'):        
            self.cdms_axes = []

        for ivar_number in range(self.na_file_obj.NIV):
            self.cdms_axes.append(self._convertNAIndVarToCdmsAxis(ivar_number))


    def _convertNAIndVarToCdmsAxis(self, ivar_number):
        """
        Creates a cdms axis from a NASA Ames independent variable.
        """
        if self.na_file_obj._normalized_X == False:   self.na_file_obj._normalizeIndVars()

        if self.na_file_obj.NIV == 1:
            array = self.na_file_obj.X
        else:
            array = self.na_file_obj.X[ivar_number]

        axis = cdms.createAxis(array)
        axis.id = axis.name = axis.long_name = self.na_file_obj.XNAME[ivar_number]
        (var_name, units) = self.na_file_obj.getIndependentVariable(ivar_number)
	
        # Sort units etc
        if units:   axis.units = units
        if len(var_name) < max_id_length:
            axis.id = safe_nc_id.sub("_", var_name).lower()
        else:
            axis.id = "naIndVariable_%s" % (ivar_number)

        if units: axis.units = units

        axis_types = ("longitude", "latitude", "level", "time")

        for axis_type in axis_types:
            if re.search(axis_type, var_name, re.IGNORECASE):
                axis.standard_name = axis.id = axis_type
                # Designate it CF-style if known axis type (e.g. axis.designateTime() etc..)
                exec "axis.designate%s()" % axis_type.title()

        # Check warning for time units pattern
        if axis.isTime() and (not hasattr(axis, "units") or not time_units_pattn.match(axis.units)):
            if self.time_units == None:
                time_units_input = "I WON'T MATCH"

                while time_units_input != "" and not time_units_pattn.match(time_units_input):
                    message = time_units_warning_message			    
                    if self.time_warning == True:
                        log.debug(message)
                        time_units_input = raw_input("Please insert your time unit string here (or leave blank):").strip()
                    else: 
                        time_units_input = ""

                self.output_message.append(message)
                self.time_units = time_units_input

	    axis.units = self.time_units
            axis.long_name = axis.name = "time (%s)" % self.time_units

        if not hasattr(axis, "units") or axis.units == None:  
            if units:
                axis.units = units	
            else:
                axis.units = "Not known"

        return axis



# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

from setuptools import setup, find_packages

setup(
      name='nappy',
      version='1.1.4',

      description = 'NASA Ames Processing in Python',
      long_description = "A python package for reading/writing NASA Ames files, writing NASA Ames-style CSV files and converting to/from NetCDF (if CDMS enabled).",
      keywords = 'Python CSV NASA Ames NetCDF convert CDMS',
      author = 'Ag Stephens',
      author_email = 'ag.stephens@stfc.ac.uk',
      url = 'https://github.com/cedadev/nappy',

      # We make the package non-zip_safe so that we don't need to
      # change the way ini files are read too much.
      packages = find_packages('lib', exclude=['tests', 'tests.*']),
      package_dir={'': 'lib'},

      include_package_data = True,
      zip_safe = False,

      tests_require = ['nose'],
      test_suite = 'nose.collector',

      entry_points = '''
      [console_scripts]
      na2nc = nappy.script.na2nc:na2nc
      nc2na = nappy.script.nc2na:nc2na
      nc2csv = nappy.script.nc2csv:nc2csv
      ''',
      
)
